"""Helper functions for the class pairwise alignment
Creates its own logger object when imported.

Functions:

"""

import numpy as np
import skimage.transform as smtf
import logging
# import matplotlib.pyplot as plt

from . import inout

logger = logging.getLogger(__name__)


####################Third level helper functions####################
def get_overlapping_region(tile_1, tile_2,
                           overlap_ind_x, overlap_ind_y, direction):
    """Given the overlap indexes calculate the overlap in direction.

    Parameters:
    -----------

    tile_1: np.array
        2D or 3D numpy array representing an image.
    tile_2: np.array
        2D or 3D numpy array representing an image.
    overlap_ind_x: int
        The difference in pixels between the tile corners in the x/column direction.
    overlap_ind_y: int
        The difference in pixels between the tile corners in the y/row direction.
    direction: str
        Valid values: 'left' or 'top'. Denotes if
        the two tiles neighbour on the right and left or
        at the bottom and top.

    Returns:
    --------

    overlap_1: np.array
        2D or 3D numpy arrays. The parts of each image that overlap.
    overlap_2: np.array
        2D or 3D numpy arrays. The parts of each image that overlap.

    """
    if (direction == 'left' and tile_1.ndim == 2):
        # Adjust y overlap on left:
        if overlap_ind_y == 0:
            overlap_1 = tile_1[:, -overlap_ind_x:]
            overlap_2 = tile_2[:, :overlap_ind_x]
        elif overlap_ind_y < 0:
            # For negative difference in y index select overlap:
            overlap_1 = tile_1[-overlap_ind_y:, -overlap_ind_x:]
            overlap_2 = tile_2[:overlap_ind_y, :overlap_ind_x]
        else:
            # For positive y index select overlap
            overlap_1 = tile_1[:-overlap_ind_y, -overlap_ind_x:]
            overlap_2 = tile_2[overlap_ind_y:, :overlap_ind_x]
    elif (direction == 'left' and tile_1.ndim == 3):
        # Adjust y overlap on left:
        if overlap_ind_y == 0:
            overlap_1 = tile_1[:, :, -overlap_ind_x:]
            overlap_2 = tile_2[:, :, :overlap_ind_x]
        elif overlap_ind_y < 0:
            # For negative difference in y index select overlap:
            overlap_1 = tile_1[:, -overlap_ind_y:, -overlap_ind_x:]
            overlap_2 = tile_2[:, :overlap_ind_y, :overlap_ind_x]
        else:
            # For positive y index select overlap
            overlap_1 = tile_1[:, :-overlap_ind_y, -overlap_ind_x:]
            overlap_2 = tile_2[:, overlap_ind_y:, :overlap_ind_x]
    elif (direction == 'top' and tile_1.ndim == 2):
        # Adjust x overlap:
        if overlap_ind_x == 0:
            overlap_1 = tile_1[-overlap_ind_y:, :]
            overlap_2 = tile_2[:overlap_ind_y, :]
        elif overlap_ind_x < 0:
            # For negative difference in x index select overlap:
            overlap_1 = tile_1[-overlap_ind_y:, -overlap_ind_x:]
            overlap_2 = tile_2[:overlap_ind_y, :overlap_ind_x]
        else:
            # For positive x index select overlap
            overlap_1 = tile_1[-overlap_ind_y:, :-overlap_ind_x]
            overlap_2 = tile_2[:overlap_ind_y, overlap_ind_x:]
    elif (direction == 'top' and tile_1.ndim == 3):
        # Adjust x overlap:
        if overlap_ind_x == 0:
            overlap_1 = tile_1[:, -overlap_ind_y:, :]
            overlap_2 = tile_2[:, :overlap_ind_y, :]
        elif overlap_ind_x < 0:
            # For negative difference in x index select overlap:
            overlap_1 = tile_1[:, -overlap_ind_y:, -overlap_ind_x:]
            overlap_2 = tile_2[:, :overlap_ind_y, :overlap_ind_x]
        else:
            # For positive x index select overlap
            overlap_1 = tile_1[:, -overlap_ind_y:, :-overlap_ind_x]
            overlap_2 = tile_2[:, :overlap_ind_y, overlap_ind_x:]
    else:
        logger.warning("The overlapping region could not be retrieved\n"
                       + "The overlap direction and dimension of the image do not match any known case\n"
                       + "Direction: {} Dimension: {}".format(direction,
                                                              tile_1.ndim))

    return overlap_1, overlap_2


def calculate_PCM(pic_a, pic_b):
    """Calculate the phase correlation matrix of pic_a and pic_b.

    Parameters:
    -----------

    pic_a: np.array
        First picture
    pic_b: np.array
        Second picture

    Returns:
    --------

    r: np.array
        Normalized PCM matrix

    Notes:
    ------
        pic_a and pic_b should have the same size


    """
    # From: http://stackoverflow.com/questions/2771021/is-there-an-image-phase-correlation-library-available-for-python
    # And: https://github.com/michaelting/Phase_Correlation/blob/master/phase_corr.py
    # Calculate phase correlation matrix
    # Calculate nD fourier transform of first image
    G_a = np.fft.fftn(pic_a)
    # Calculate nD fourier transform of second image
    G_b = np.fft.fftn(pic_b)
    conj_b = np.ma.conjugate(G_b)  # Take the complex conjunggate of G_b
    R_raw = G_a * conj_b  # Multiply G_a by the complex conjugate of G_b
    R = R_raw / np.absolute(R_raw)  # Normalize elementwise
    # Inverse fourier transform to get correlation, and use only the real part
    r = np.fft.ifftn(R).real
    # inout.plot_3D(r)
    #plt.figure()
    #plt.imshow(r, 'gray', interpolation = 'none')
    #plt.show()

    return r


def calculate_PCM_method2(pic_a, pic_b):
    """Calculate the phase correlation matrix of pic_a and pic_b.

    Method 2 differs from calculate_PCM in that it calculates the PCM
    per, layer in the image therefore this function is only applicable
    to 3D images/complete z-stacks.
    In fact it is only used when in pairwisesingle.py if method ==
    'calculate per layer'.

    Parameters:
    -----------

    pic_a: np.array
        First picture
    pic_b: np.array
        Second picture

    Returns:
    --------

    r: np.array
        Normalized PCM matrix

    Notes:
    ------
        pic_a and pic_b should have the same size
    """
    r       = np.empty(pic_a.shape)
    for i in range(len(pic_a[:, 0, 0])):
        r[i, :, :] = calculate_PCM(pic_a[i, :, :], pic_b[i, :, :])
    return r


def calc_translated_pics(trans, overlap1, overlap2,
                         round_size=False):
    """Calculate wich part of 2 pictures overlap after translation.

    Translates overlap2 and returns the parts of overlap1 and
    overlap2 that overlap after translation. Parts that do not
    overlap are cropped. The alternative would be to fill the parts
    that do not overlap with zeros, but this seems to mess up the
    covariance calculation more than cropping the pictures.
    Here, one problem is that we may crop unnecessarily much
    because the cropping is done on the overlap and not on the whole
    tile.

    Parameters:
    -----------

    trans: np.array
        1 by 2 np-array with y and x translation.
    overlap1: np.array
        Image
    overlap2: np.array
        Image that is supposed to overlap with overlap1
    round_size: bool
        If True the final image sizes will be
        rounded to the nearest integer before warping the
        images. (Default: False)

    Returns:
    --------

    shifted_a: np.array
        Overlap a shifted and cropped to the same size as shifted_b
    shifted_b: np.array
        Overlap b shifted and cropped to the same size as shifted_a
    """
    # Pass to calc_translated_pics_3D if there are more then 2 dimensions
    if overlap1.ndim > 2:
        return calc_translated_pics_3D(trans, overlap1, overlap2,
                                       round_size=round_size)

    # Empty transformation matrix
    trans_matrix = np.eye(3)

    # Size after shift
    y_size, x_size = overlap2.shape
    # Always make new size smaller
    y_size -= abs(trans[0])
    x_size -= abs(trans[1])

    if round_size:
        y_size = int(round(y_size))
        x_size = int(round(x_size))

    # Transform a with transform to cut  on left and top
    if (trans[0] < 0):
        trans_matrix[1, 2] = abs(trans[0])
    if (trans[1] < 0):
        trans_matrix[0, 2] = abs(trans[1])
    # Or transform a with empty matrix (0,0) to make smaller on right
    logger.debug("x and y size: {} {}".format(x_size, y_size))
    shifted_a = smtf.warp(overlap1, trans_matrix,
                          output_shape = (y_size, x_size))
    # Reset transformation matrix
    trans_matrix = np.eye(3)

    # Set y transformation
    if trans[0] > 0:
        trans_matrix[1, 2] = trans[0]
    # Set x transformation
    if trans[1] > 0:
        trans_matrix[0, 2] = trans[1]

    # Transform b with empty matrix (0,0) to give it the right size
    shifted_b = smtf.warp(overlap2, trans_matrix,
                          output_shape=(y_size, x_size))

    return shifted_a, shifted_b


def calc_translated_pics_3D(trans, overlap1, overlap2,
                            round_size=False):
    """Calculate wich part of two 3D pictures overlap after translation.

    Translates overlap2 and returns the parts of overlap1 and
    overlap2 that overlap after translation.
    This function works in 3D but does not translate in z.
    Parts that do not overlap are cropped. The alternative would be to
    fill the parts that do not overlap with zeros, but this seems to
    mess up the covariance calculation more than cropping the pictures.
    Here, one problem is that we may crop unnecessarily much
    because the cropping is done on the overlap and not on the whole
    tile.

    Parameters:
    -----------

    Parameters:
    -----------

    trans: np.array
        1 by 2 np-array with y and x translation.
    overlap1: np.array
        Image
    overlap2: np.array
        Image that is supposed to overlap with overlap1
    round_size: bool
        If True the final image sizes will be
        rounded to the nearest integer before warping the
        images. (Default: False)

    Returns:
    --------

    shifted_a: np.array
        Overlap a shifted and cropped to the same size as shifted_b
    shifted_b: np.array
        Overlap b shifted and cropped to the same size as shifted_a
    """

    logger.debug("Trans: {}".format(trans))
    # Size after shift
    z_size, y_size, x_size = overlap2.shape
    # Always make new size smaller
    y_size -= abs(trans[1])
    x_size -= abs(trans[2])

    if round_size:
        y_size = int(round(y_size))
        x_size = int(round(x_size))

    # Do x and y translation for each layer:
    # Empty transformation matrix
    trans_matrix = np.eye(3)

    # Transform a with transform to cut  on left and top
    if (trans[1] < 0):
        # y trans
        trans_matrix[1, 2] = abs(trans[1])
    if (trans[2] < 0):
        # x trans
        trans_matrix[0, 2] = abs(trans[2])
    # Or transform a with empty matrix (0,0) to make smaller on right
    logger.debug("Trans, z, x and y size: {} {} {} {}"
                 .format(trans, z_size, x_size, y_size))
    logger.debug("Matrix: {}"
                 .format(trans_matrix))
    shifted_a = np.empty((z_size, y_size, x_size))
    logger.debug("Shape overlap 1, after z transform: {}"
                 .format(overlap1.shape))
    for i in range(overlap1.shape[0]):
        shifted_a[i, :, :] = smtf.warp(overlap1[i, :, :], trans_matrix,
                                       output_shape=(y_size, x_size))
        # plt.figure()
        # plt.imshow(shifted_a[i,:,:])
        # plt.show()

    # Do x and y transform in each layer
    # Reset transformation matrix
    trans_matrix = np.eye(3)

    # Set y transformation
    if trans[1] > 0:
        trans_matrix[1, 2] = trans[1]
    # Set x transformation
    if trans[2] > 0:
        trans_matrix[0, 2] = trans[2]

    shifted_b = np.empty((z_size, y_size, x_size))
    # Transform b with empty matrix (0,0) to give it the right size
    for i in range(overlap2.shape[0]):
        shifted_b[i, :, :] = smtf.warp(overlap2[i, :, :], trans_matrix,
                                       output_shape=(y_size, x_size))
    return shifted_a, shifted_b


def xcov_nd(pic_1, pic_2):
    """Calculate the normalized cross covariance of pic_1 and pic_2.

    Parameters:
    -----------

    pic_1: np.array
        Numpy array representing a gray value picture, same size as pic_2
    pic_2: np.array 
        Numpy array representing a gray value picture, same size as pic_1

    Returns:
    --------

    : float
         The normalized crosscovariance of pic_1 and pic_2 (coVar / (stDev1 * stDev2))
    monocolor: bool   
        True if the one or both of the pictures contained only one color.
    """
    # Converted from Java code used in Preibisch plugin:
    # https://github.com/tomka/imglib/blob/master/mpicbg/imglib/algorithm/fft/PhaseCorrelation.java
    # Alternative use: scipy.signal.correlate2d
    monocolor = False
    avg1 = np.mean(pic_1)
    avg2 = np.mean(pic_2)
    dist1M = pic_1 - avg1
    dist2M = pic_2 - avg2

    # Calculate var and co variance for the whole picture, per pixel, elementwise multiplication
    coVarMatrix = dist1M * dist2M
    var1Matrix = dist1M * dist1M
    var2Matrix = dist2M * dist2M

    # Take the average
    coVar   = np.mean(coVarMatrix)
    var1    = np.mean(var1Matrix)
    var2    = np.mean(var2Matrix)

    stDev1 = np.sqrt(var1)
    stDev2 = np.sqrt(var2)

    # All pixels had the same color....
    if (stDev1 == 0 or stDev2 == 0):
        monocolor = True
        if (stDev1 == stDev2 and avg1 == avg2):
            return (1, monocolor)
        else:
            return (0, monocolor)

    # Compute correlation coeffienct
    return (coVar / (stDev1 * stDev2), monocolor)


def display_overlap(overlap1, overlap2, best_trans, best_cov,
                    plot_order):
    """Plot two pictures and show how they overlap.

    This function will only plot when allow_plot is True and in inout
    matplotlib is imported and plot_available is True.

    Parameters:
    -----------

    overlap1: np.array
        2d np-array representing an image
    overlap2: np.array
        2d np-array representing an image
    best_trans: np.array
        1 by 2 np array with an y and x value for the best transition
    best_cov: float
        Number indicating the goodness of the overlap between the two pictures
    plot_order: list
        The order in wich subplots should be placed if when plotting overlap1 
        and 2 and overlap rgb
    """
    allow_plot = False
    if allow_plot:
        logger.debug("Displaying overlap")
        logger.debug("best trans, cov: {}, {}"
                     .format(best_trans, best_cov))

        shifted_a, shifted_b = calc_translated_pics(best_trans,
                                                    overlap1, overlap2,
                                                    round_size=True)
        best_cov = xcov_nd(shifted_a, shifted_b)

        # overlap_rgb         = np.zeros((shifted_a.shape[0], shifted_a.shape[1],  3))
        # overlap_rgb[:,:,0]  = smex.rescale_intensity(shifted_a)
        # overlap_rgb[:,:,1]  = smex.rescale_intensity(shifted_b)

        title_text = "Pairwise overlap, cov: {}".format(best_cov)

        # inout.display_tiles([shifted_a,
        #                         shifted_b,
        #                         overlap_rgb],
        #                         plot_order, fig_nr = "overlap check",
        #                         maximize = False, main_title = title_text,
        #                         block = True)
        overlap_sum = shifted_a + shifted_b
        inout.display_tiles([shifted_a,
                             shifted_b,
                             overlap_sum],
                            plot_order, fig_nr="overlap check",
                            maximize=False, main_title=title_text,
                            block=True)
        # if shifted_a.ndim == 3:
        #    inout.plot_3D(overlap_sum)
    else:
        return None



import pickle
import numpy as np
from skimage import io, img_as_float
from .dots_calling import thr_calculator
from .filtering import smFISH_filtering, nuclei_filtering



def filtering_and_counting(fpath_img_to_filter,filtered_png_img_gene_dirs,filtered_img_gene_dirs,
        counting_gene_dirs, illumination_correction=False ,plane_keep=None, min_distance=5, stringency=0,
        skip_genes_counting=None,skip_tags_counting=None):
    """
    Function used to clean the images and to count the smFISH dots.
    It is designed to process in parallel all the tmp file images
    stored as numpy arrays after conversion from the microscope format.
    
    Parameters:
    ------------

    fpath_img_to_filter: str 
        path to the file to process
    filtered_png_img_gene_dirs: list 
        list of the paths of the directories where the filtered images as are
        saved as pngs.
    filtered_img_gene_dirs: list 
        list of the paths of the directories where the filtered images are saved 
        as .npy.
    counting_gene_dirs: list 
        list of the paths of the directories where the countings of the filtered
        images are saved.
    illumination_correction: bool 
        if True the illumination correction is run on the dataset.
    plane_keep: list 
        start and end point of the z-planes to keep. Default None
        keep all the planes (ex. [2,-3]).
    min_distance: int 
        minimum distance between dots.
    stringency: int 
        stringency use to select the threshold used for counting.
    skip_genes_counting: list 
        list of the genes to skip for counting count.
    skip_tags_counting: list 
        list of the tags inside the genes/stainings name to avoid to count.
    """
    
    # Get infos from file name
    fname_split = fpath_img_to_filter.split('/')[-1].split('_')
    experiment_name = fname_split[0]
    hyb = fname_split[1]
    gene = fname_split[2]
    pos = fname_split[4].split('.')[0]
    
    
    # Load the image to process
    img_stack = np.load(fpath_img_to_filter) # image is np.uint16
    img_stack = img_as_float(img_stack)

    # Remove extra planes. As it is for now this step is mainly for single image
    # usage. I will include the automatic excess planes remove function to use
    # for large scale image analysis later on
    if isinstance(plane_keep,list):
        img_stack = img_stack[plane_keep[0]:plane_keep[1],:,:]


    # Filtering image according to gene
    if gene in skip_genes_counting or [tag for tag in skip_tags_counting if tag in gene]:
        # Remove the background from the nuclei
        img_filtered = nuclei_filtering(img_stack)
        counting_dict = None

    else:
        # Remove background and enhance smFISH signal
        img_filtered=smFISH_filtering(img_stack)

        # Count the dots in the whole image
        counting_dict=thr_calculator(img_filtered,min_distance,stringency)


    # Convert image to uint16
    # Clip the values above 1
    img_filtered[img_filtered>1] = 1
    # Scale to the max of the uint16
    img_filtered *= np.iinfo(np.uint16).max
    # Round and convert to integer
    img_filtered = np.uint16(np.rint(img_filtered))
    
    # Save images and dictionary
    # This part may be removed from the function in case we will run
    # temporary storage in RAM in order to reduce i/o to the common
    # HD of the cluster
    # Identify the directory for storing the images and the counting
    img_saving_dir_npy=[saving_dir for saving_dir in filtered_img_gene_dirs if gene in saving_dir.split('/')[-2] ][0]

    img_saving_dir_png=[saving_dir for saving_dir in filtered_png_img_gene_dirs if gene in saving_dir.split('/')[-2] ][0]
    

    # Save the images and the counting if performed
    fname_png = img_saving_dir_png+experiment_name+'_'+hyb+'_'+gene+'_'+'pos_'+pos+'.png'
    io.imsave(fname_png,img_filtered)

    fname_npy = img_saving_dir_npy+experiment_name+'_'+hyb+'_'+gene+'_'+'pos_'+pos+'.npy'
    np.save(fname_npy,img_filtered,allow_pickle=False)


    if counting_dict:
        # may missing if I don't want the counting
        counting_saving_dir=[saving_dir for saving_dir in counting_gene_dirs if gene in saving_dir.split('/')[-2] ][0]
        fname = counting_saving_dir+experiment_name+'_'+hyb+'_'+gene+'_'+'pos_'+pos+'.pkl'
        pickle.dump(counting_dict,open(fname,'wb'))
           
    return


def filtering_and_counting_experiment(fpath_img_to_filter,filtered_dir_path,
                                      counting_dir_path,exp_name,add_slash,plane_keep=None,
                                      min_distance=5,stringency=0):
    
    """
    Function to filter and count dots in the images generated from a small
    experiment.

    Parameters:
    ------------

    fpath_img_to_filter: str 
        path to the file to process.
    filtered_dir_path: list 
        list of the paths of the directories where the filtered images are saved.
    counting_dir_path: list 
        list of the paths of the directories where the counting of filtered 
        images are stored.
    exp_name: str 
        name of the experiment to process.
    plane_keep: list 
        start and end point of the z-planes to keep. Default None
        keep all the planes (ex. [2,-3]).
    min_distance: int 
        minimum distance between dots.
    stringency: int 
        stringency use to select the threshold used for counting.
    
    """

    # Load the image to process
    img_stack = np.load(fpath_img_to_filter) # image is np.uint16
    img_stack = img_as_float(img_stack)

    # Remove extra planes. As it is for now this step is mainly for single image
    # usage. I will include the automatic excess planes remove function to use
    # for large scale image analysis later on
    if plane_keep:
        img_stack = img_stack[plane_keep[0]:plane_keep[1],:,:]

    channel = fpath_img_to_filter.split(add_slash)[-1].split('_')[-3]
    fov = fpath_img_to_filter.split(add_slash)[-1].split('_')[-1].split('.')[0]
    
    not_counting=['Nuclei','Dapi','DAPI']
    
    # Filtering image according to gene
    if channel in not_counting or '_IF' in channel or channel == 'polyA':
        # Remove the background from the nuclei
        img_filtered = nuclei_filtering(img_stack)
        counting_dict = None

    else:
        # Remove background and enhance smFISH signal
        img_filtered=smFISH_filtering(img_stack)

        # Count the dots in the whole image
        counting_dict=thr_calculator(img_filtered,min_distance,stringency)
    
    
    # Convert image to uint16
    # Clip the values above 1
    img_filtered[img_filtered>1] = 1
    # Scale to the max of the uint16
    img_filtered *= np.iinfo(np.uint16).max
    # Round and convert to integer
    img_filtered = np.uint16(np.rint(img_filtered))
    
    fname = fpath_img_to_filter.split(add_slash)[-1][:-4]

    fname_path_png = filtered_dir_path+add_slash+exp_name+'_'+fname+'.png'
    io.imsave(fname_path_png,img_filtered)
    if counting_dict:
            fname_path_pkl = counting_dir_path+add_slash+exp_name+'_'+fname+'.pkl'
            pickle.dump(counting_dict,open(fname_path_pkl,'wb'))
           
    return


def filtering_and_counting_ill_correction(fpath_img_to_filter,illumination_function, filtered_png_img_gene_dirs,filtered_img_gene_dirs,
        counting_gene_dirs, illumination_correction=False ,plane_keep=None, min_distance=5, stringency=0,
        skip_genes_counting=None,skip_tags_counting=None):
    """
    Function used to clean the images and to count the smFISH dots.
    Designed to work in parallel processing all the tmp file images
    stored as numpy arrays after conversion from the microscope format.
    
    Parameters:
    ------------
    
    fpath_img_to_filter: str 
        path to the file to process.
    illumination_function: np.array float64 
         illumination function.
    filtered_png_img_gene_dirs: list 
        list of the paths of the directories where the filtered images as are
        saved as pngs.
    filtered_img_gene_dirs: list 
        list of the paths of the directories where the filtered images are saved 
        as .npy.
    counting_gene_dirs: list 
        list of the paths of the directories where the countings of the filtered
        images are saved.
    illumination_correction: bool 
        if True the illumination correction is run on the dataset.
    plane_keep: list 
        start and end point of the z-planes to keep. Default None
        keep all the planes (ex. [2,-3]).
    min_distance: int 
        minimum distance between dots.
    stringency: int 
        stringency use to select the threshold used for counting.
    skip_genes_counting: list 
        list of the genes to skip for counting count.
    skip_tags_counting: list 
        list of the tags inside the genes/stainings name to avoid to count.

    """
    
    # Get infos from file name
    fname_split = fpath_img_to_filter.split('/')[-1].split('_')
    experiment_name = fname_split[0]
    hyb = fname_split[1]
    gene = fname_split[2]
    pos = fname_split[4].split('.')[0]
    
    
    # Load the image to process
    img_stack = np.load(fpath_img_to_filter) # image is np.uint16
    img_stack = img_as_float(img_stack)

    # Remove extra planes. As it is for now this step is mainly for single image
    # usage. I will include the automatic excess planes remove function to use
    # for large scale image analysis later on
    if isinstance(plane_keep,list):
        img_stack = img_stack[plane_keep[0]:plane_keep[1],:,:]


    # Correct for illumination
    img_stack = img_stack/illumination_function

    # Filtering image according to gene
    if gene in skip_genes_counting or [tag for tag in skip_tags_counting if tag in gene]:
        # Remove the background from the nuclei
        img_filtered = nuclei_filtering(img_stack)
        counting_dict = None

    else:
        # Remove background and enhance smFISH signal
        img_filtered=smFISH_filtering(img_stack)

        # Count the dots in the whole image
        counting_dict=thr_calculator(img_filtered,min_distance,stringency)
        
    # Convert image to uint16
    # Clip the values above 1
    img_filtered[img_filtered>1] = 1
    # Scale to the max of the uint16
    img_filtered *= np.iinfo(np.uint16).max
    # Round and convert to integer
    img_filtered = np.uint16(np.rint(img_filtered))
    
    # Save images and dictionary
    # This part may be removed from the function in case we will run
    # temporary storage in RAM in order to reduce i/o to the common
    # HD of the cluster
    # Identify the directory for storing the images and the counting
    img_saving_dir_npy=[saving_dir for saving_dir in filtered_img_gene_dirs if gene in saving_dir.split('/')[-2] ][0]

    img_saving_dir_png=[saving_dir for saving_dir in filtered_png_img_gene_dirs if gene in saving_dir.split('/')[-2] ][0]

    

    # Save the images and the counting if performed
    fname_png = img_saving_dir_png+experiment_name+'_'+hyb+'_'+gene+'_'+'pos_'+pos+'.png'
    io.imsave(fname_png,img_filtered)

    fname_npy = img_saving_dir_npy+experiment_name+'_'+hyb+'_'+gene+'_'+'pos_'+pos+'.npy'
    np.save(fname_npy,img_filtered,allow_pickle=False)


    if counting_dict:
        # may missing if I don't want the counting
        counting_saving_dir=[saving_dir for saving_dir in counting_gene_dirs if gene in saving_dir.split('/')[-2] ][0]
        fname = counting_saving_dir+experiment_name+'_'+hyb+'_'+gene+'_'+'pos_'+pos+'.pkl'
        pickle.dump(counting_dict,open(fname,'wb'))
           
    return


"""
Collection of utility functions.

"""

import os
import logging
from logging.handlers import RotatingFileHandler
import glob
import time
import yaml
import ruamel.yaml
import codecs
from collections import OrderedDict
import nd2reader as nd2
import numpy as np
from skimage import img_as_float
import platform


##################################Logging###############################
"""
The following lines will be executed when this file is imported, this code
initializes a logger, but does not use the logger. To print the output from
the logger run the function init_file_logger and/or init_console_logger, to save
output to a file or send the output to the stderr stream respectively.
"""
# Redirect warnings to logger
logging.captureWarnings(True)

# Create logger
logger = logging.getLogger()
logger.setLevel(logging.DEBUG)


def init_file_logger(log_path, maxBytes=0, backupCount=0):
    """
    Send the logging output to a file.

    The logs are placed in a directory called "logs", if this directory
    is not found at the location denoted by log_path, it will be made.
    On each run of the program a new log file will be produced.
    max_bytes and backup_count are passed to RotatingFileHandler
    directly. So a new file will be produced when the filesize
    reaches max_bytes. If either of max_bytes or backup_count is zero,
    rollover never occurs and everything will be logged in one file.
    
    Parameters:
    -----------

    log_path: str 
        Full path to the directory where the log directory is/should go.
    maxBytes: int 
        Maximum size in bytes of a single log file, a new log file will be 
        started when this size is reached. If zero, all logging will be 
        written to one file. (Default: 0)
    backup_count: int 
        The maxinum number of logging file that will be produced. If zero, 
        all logging will be written to one file. (Default: 0)
    """
    formatter_file = logging.Formatter('%(asctime)s - %(levelname)s - %(name)s: %(message)s')
    log_path += '_logs/'
    # To print the logging to a file
    try:
        os.stat(log_path)
    except:
        os.mkdir(log_path)
        os.chmod(log_path, 0o777)

    date_tag = time.strftime("%y%m%d_%H_%M_%S")
    fh = RotatingFileHandler(log_path +'%s-smFISH_Analysis.log' % date_tag,
                            mode ='w', maxBytes=maxBytes,
                            backupCount=backupCount)

    # Set logger properties
    fh.setLevel(logging.DEBUG)
    fh.setFormatter(formatter_file)
    logger.addHandler(fh)
    logger.info("Created file logger")


def init_console_logger():
    """
    Send the logging output to the stderr stream.
    After running this function the logging message will typically end up
    in your console output.
    """
    formatter_con = logging.Formatter('%(name)s: %(message)s')
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    ch.setFormatter(formatter_con)
    logger.addHandler(ch)
    logger.info("Created console logger")

########################################################################

def experimental_metadata_parser(hyb_dir):
    """
    Parse the yaml file containing all the metadata of the experiment

    The file must be located inside the experimental folder.

    Parameters:
    -----------

    hyb_dir: str
        Path to the .yaml file containing the metadata of the experiment

    Returns:
    -----------

    experiment_infos: dict 
        Dictionary with the information on the experiment.
    HybridizationInfos: dict 
        Dictionary with the information on the hybridization.
    converted_positions: dict 
        Dictionary with the coords of the images for all hybridization. 
        The coords are a list of floats
    microscope_parameters: dict 
        Dictionary with the microscope parameters for all hybridization

    """
    metadata_file_name = hyb_dir+'Experimental_metadata.yaml'
    logger.info("Parsing metadata from file: {}"
                .format(metadata_file_name))
    with open(metadata_file_name, 'r') as stream:
        try:
            docs=yaml.load(stream)
        except yaml.YAMLError as exc:
            print(exc)

    experiment_infos = docs['ExperimentInfos']

    image_properties = docs['ImageProperties']
    hybridizations_infos = docs['HybridizationsInfos']
    Positions = docs['TilesPositions']
    microscope_parameters = docs['MicroscopeParameters']

    # Dictionary that will contain the coords after been coverted to float
    converted_positions={}
    # Convert the positions from list of string to list of floats
    for hyb, coords_dict in Positions.items():
        sub_dict = {}
        for pos, coords in coords_dict.items():
            coords = [float(x) for x in coords.split(',')]
            sub_dict[pos] = coords
        converted_positions[hyb] = sub_dict

    return experiment_infos,image_properties, hybridizations_infos,\
           converted_positions, microscope_parameters



def filtering_raw_counting_config_parser(hyb_dir):
    """
    Parse the yaml file containing all configurations for running the analysis

    The file must be located inside the experimental folder.

    Parameters:
    -----------

    hyb_dir: str 
        Path to the .yaml file containing the metadata of the experiment

    Returns:
    -----------

    config_parameters: dict 
        Dictionary with all the configuration parameters
    """
    configuration_file_name = hyb_dir+'Filtering_raw_counting.config.yaml'
    logger.info("Parsing metadata from file: {}"
            .format(configuration_file_name))
    with open(configuration_file_name, 'r') as stream:
        try:
            config_parameters=yaml.load(stream)
        except yaml.YAMLError as exc:
            print(exc)

    return config_parameters


def general_yaml_parser(file_path):
    """
    Parse a general yaml file and return the dictionary with all the 
    content

    The file must be located inside the experimental folder.

    Parameters:
    -----------

    file_path: str 
        Path to the .yaml file containing the metadata of the experiment

    Returns:
    -----------

    parameters: dict 
        Dictionary with all the configuration parameters
    """
    
    logger.info("Parsing metadata from file: {}"
            .format(file_path))
    with open(file_path, 'r') as stream:
        try:
            parameters=yaml.load(stream)
        except yaml.YAMLError as exc:
            print(exc)

    return parameters


def determine_os():
    """
    This function check if the system is running windows.
    and return the correct slash type to use

    Returns:
    --------

    os_windows: bool 
        True if the os is windows.
    add_slash: str 
        '\' for windows or '/' for any other system

    """

    if 'Windows' in platform.platform():
        os_windows = True
        add_slash = '\\'
    else:
        os_windows = False
        add_slash = '/'

    if os_windows:
        logger.debug('OS: Windows')
    else:
        logger.debug('OS: Linux based')
        
    return os_windows, add_slash



def check_trailing_slash(dir_path, os_windows):
    """
    This function check if there is a trailing slash at the end of a 
    directory path and add it if missing

    Paramenters:
    ------------

    dir_path= str 
        Path to the directory

    Returns:
    --------

    dir_path= str 
        Path to the directory

    """
    logger.info('Checking the trailing slash ')
    
    if os_windows:
        if dir_path[-1]=='\\':
            logger.info('trailing slash present')
        else:
            logger.info('missing trailer slash, added now')
            dir_path=dir_path+'\\'
    else:

        if dir_path[-1]=='/':
            logger.info('trailing slash present')
        else:
            logger.info('missing trailer slash, added now')
            dir_path=dir_path+'/'
    return dir_path


def create_subdirectory_tree(hyb_dir,hybridization,hybridizations_infos,processing_hyb,suffix,add_slash,
                                skip_tags=None,skip_genes=None,analysis_name=None):
    """
    Function that creates the directory tree where to save the
    temporary data.
    
    Parameters:
    -----------

    hyb_dir: str
        Path of the hyb to process
    hybridization: str 
        Name of the hybridization to process (ex. Hybridization2)
    hybridizations_infos: dict 
        Dictionary containing the hybridizations info parsed from the 
        Experimental_metadata.yaml file
    processing_hyb: str 
        Name of the processing experiment (ex. EXP-17-BP3597_hyb2)
    suffix: str 
        Suffix to add to the folder with useful description (ex. tmp)
    add_slash: str 
        '\\' for win and '/' for linux
    skip_tags: list 
        tags that won't be processed (ex. _IF)
    skip_genes list
         list of genes to skip
    analysis_name: str 
        Name of the analysis run

    Returns:
    ---------

    sufx_dir_path: str
        Path of the sufx directory of the processed hybridization
    sufx_gene_dirs: list 
        List of the paths of the sufx directory for the genes to process
    """

    logger.info('create {} directory'.format(suffix))

    gene_list = list(hybridizations_infos[hybridization].keys())
    # logger.debug('gene list: {}'.format(gene_list))
    sufx_gene_dirs = []
    if analysis_name: 
        # Create sufx directory
        sufx_dir_path = hyb_dir+analysis_name+'_'+processing_hyb+'_'+suffix+add_slash
        logger.debug('create {} directory'.format(suffix))
    else:
        # Create sufx directory
        sufx_dir_path = hyb_dir+processing_hyb+'_'+suffix+add_slash
    try:
        os.stat(sufx_dir_path)
    except:
        os.mkdir(sufx_dir_path)
        os.chmod(sufx_dir_path,0o777)
    
    
    if skip_genes:
        gene_list = [gene for gene in gene_list if gene not in skip_genes]
        
    if skip_tags:
        gene_list = [gene for tag in skip_tags for gene in gene_list if tag not in gene]

    for gene in gene_list:         
        if analysis_name:
            sufx_gene_dir_path = sufx_dir_path+analysis_name+'_'+processing_hyb+'_'+ gene+'_'+suffix+add_slash
            sufx_gene_dirs.append(sufx_gene_dir_path)
        else:
            sufx_gene_dir_path = sufx_dir_path +processing_hyb+'_'+ gene +'_'+suffix+add_slash
            sufx_gene_dirs.append(sufx_gene_dir_path)
        try:
            os.stat(sufx_gene_dir_path)
        except:
            os.mkdir(sufx_gene_dir_path)
            os.chmod(sufx_gene_dir_path,0o777)
    return sufx_dir_path, sufx_gene_dirs


def identify_nodes(client):
    """
    Function used to determine the address of the nodes in order to 
    better split the work

    Parameters:
    -----------

    client: dask.obj 
        Dask.distributed client.
    
    Returns:
    -----------

    node_addresses: OrderedDict
        Ordered dictionary. The keys are the addresses of the nodes and the 
        items are the full addresses of teh workers of a specific node.
    
    """
    logger.info('Determine the tcp addresses of the workers')
    
    # Gather the addresse of all the instantiated workers
    client_infos = client.scheduler_info()
    workers_addresses = client_infos['workers'].keys()
    
    # Isolate the tcp address of the nodes
    nodes_addresses = OrderedDict()
    nodes_comparison_list = []
    for address in workers_addresses:
        address_split = address.split(':')
        node_address = address_split[1].split('//')[-1]
        final_digits = address_split[-1]   
        if node_address in nodes_comparison_list:
            nodes_addresses['tcp://'+node_address][final_digits]=address
        else:
            nodes_comparison_list.append(node_address)
            nodes_addresses['tcp://'+node_address]={}
            nodes_addresses['tcp://'+node_address][final_digits]=address
    
    return nodes_addresses


def combine_gene_pos(hybridizations_infos,converted_positions,hybridization):
  

    """
    Gather info about the imaging at each hybridization.

    This function creates a dictionary where for each hybridization
    are shown the genes and number of positions imaged for each
    gene. This function will be useful to created distribution lists
    for running parallel processing of the datasets.

    Parameters:
    -----------

    hybridizations_infos: dict
        Dictionary with parsed Hybridizations metadata
    converted_positions: dict 
        Dictionary with the coords of the images for all hybridization. 
        The coords are a list of floats
    hybridization: str
        Selected hybridization to process

    Returns:
    -----------

    genes_and_positions: dict 
        Dictionary where for each hybridization, the genes and number of 
        positions imaged for each gene are showed.
    """
    
    genes_and_positions=dict()

    for gene in hybridizations_infos[hybridization].keys():
            genes_and_positions[gene] = list(converted_positions[hybridization].keys())
    
    return genes_and_positions


def partial_image_mean(img_paths):

    """
    Helper function used to calculate the mean of a set of images. It runs on a
    worker and help parallel image processing
    
    Parameters:
    -----------

    img_paths: list 
        List of paths to the images saved as *.npy


    Returns:
    -----------

    ImgMean: np.array 
        Array storing the calculated image mean

    """

    ImgMean = None
    for img_path in img_paths:
        img_stack = np.load(img_path)
        img_stack = img_as_float(img_stack)
        if ImgMean is None:
                ImgMean = img_stack
        else:
            ImgMean = (ImgMean + img_stack)/2.0
    return ImgMean

def list_chunking(list_to_chunk,num_chunks):

    """
    Helper function used to chunk a list in a number of sublists equal to
    num_chunks

    Parameters:
    -----------

    list_to_chunk: list 
        List to be chunked
    num_chunks: int 
        Number of sublists to obtain


    Returns:
    -----------

    chunked_list: list 
        List containing the chunked lists
    """
    size = np.int(len(list_to_chunk)/num_chunks)
    chunked_list = [list_to_chunk[i:i+size] for i  in range(0, len(list_to_chunk), size)]
    # if (len(list_to_chunk) - size*num_chunks)>0:
    #     chunked_list[-2].append(chunked_list[-1])
    #     del chunked_list[-1]
    return chunked_list


def create_single_directory(hyb_dir,gene,hybridization,processing_hyb,suffix,add_slash,
                                analysis_name=None):
    
    """
    Function used to create a subdirectory
    
    Parameters:
    -----------
    
    hyb_dir: str  
        Path to the directory of the hybridization currently processed.
    gene: str 
        Gene name to be included in the directory.
    processing_hyb: str 
        Name of the hybridization processed (ex. 'EXP-17-BP3597_hyb2').
    suffix: str 
        Extra info to add to the directory name (ex. blended).
    add_slash: str 
        Slash added according to the os.
    analysis_name: str 
        Name of the analysis associated to the folder
    
    Return
    ---------

    sufx_dir_path: str 
        Path to the created directory
    
    """
    
    logger.info('create {} directory'.format(suffix))
   
    sufx_gene_dirs = []
    if analysis_name: 
        # Create sufx directory
        sufx_dir_path = hyb_dir+analysis_name+'_'+processing_hyb+'_'+gene+'_'+suffix+add_slash
        logger.debug('create {} directory'.format(suffix))
    else:
        # Create sufx directory
        sufx_dir_path = hyb_dir+processing_hyb+'_'+gene+'_'+suffix+add_slash
    try:
        os.stat(sufx_dir_path)
    except:
        os.mkdir(sufx_dir_path)
        os.chmod(sufx_dir_path,0o777)
    
    return sufx_dir_path


def add_coords_to_yaml(folder, hyb_nr, hyb_key ='Hyb'):
    """
    Read tile number and coordinates and add them to the yaml file.
    Read the tile number and microscope coordinates for each tile from
    the microscope file called "coord_file_name" in "folder".
    Then insert them in dictionary "TilesPositions" in the yaml
    metadata file called Experimental_metadata.yaml.
    
    Parameters:
    -----------
    folder: str
        Exact path to the folder, including trailing "/"
    hyb_nr: int
        Hybridization number denoting for which hybridization we should read 
        and insert the coordinates
    hyb_key: str 
        Possible values 'Hyb' or 'Strip'. To add coordinates for stripping 
        if necessary.
    """

    # Key word to look for in the name of the coordinate file
    name_key    = 'Coords'
    hyb_key_filename     = hyb_key + str(hyb_nr)
    # Find the right file using "name_key" and hyb key
    coord_file_name = next((name for name in glob.glob(folder + '*.txt')
                           if (name_key in os.path.basename(name)) and
                           (hyb_key_filename in os.path.basename(name)))
                           , None)
    logger.info("Reading coordinates from file: {}"
                 .format(coord_file_name))

    if coord_file_name is None:
        logger.error("Coordinate file not found in folder {}, "
                      + "looking for txt-file including {} and {} in "
                      + "its name."
                      .format(folder, name_key, hyb_key_filename))

    # Load the yaml file with a special roundtrip loader, to change
    # it while keeping all comments and indentation
    metadata_file_name= folder + 'Experimental_metadata.yaml'
    with open(metadata_file_name, 'r') as stream:
        meta_data = ruamel.yaml.load(stream,
                                     ruamel.yaml.RoundTripLoader)
    stream.close()

    # Put coordinates from microscope file in TilesPositions
    if hyb_key == 'Hyb':
        cur_positions = meta_data['TilesPositions']['Hybridization'+ str(hyb_nr)]
    elif hyb_key == 'Strip':
        cur_positions = meta_data['TilesPositions']['Stripping'
                                                    + str(hyb_nr)]
    else:
        logger.warning("hyb_key not recognized, possible values are: "
                       + "'Hyb' or 'Strip'. Current value is: {}"
                       .format(hyb_key))

    # Open the coordinate file
    with codecs.open(coord_file_name, 'r', 'utf-16') as coord_file:
        # Use all lines starting with "#", assume these contain, index,
        # x and y coordinates
        for line in coord_file:
            # Use all lines starting with "#", that do not contain info
            # about "DAPI", assume these contain, index, x and y
            # coordinates
            if ('#' in line):
                # Replace the commas with dots, in case Windows
                # forced their Swedish commas on everything.
                # Clean up and split the line:
                replace_dict = {ord('#'): None, ord(','): ord('.')}
                line = line.translate(replace_dict).split('\t')
                logger.debug("Line read from coord file: {}"
                             .format(line))

                sep = ', '
                # Append the data we want to use:
                index   = int(line[0]) - 1
                x_value = float(line[1])
                y_value = float(line[2])
                z_value = float(line[3])

                cur_positions[index] = str(x_value) \
                                       + sep + str(y_value) \
                                       + sep + str(z_value)
            # Get out of the loop after we got to "Spectral Loop",
            # because we are not interested in the info that comes
            # after that.
            if ("Spectral Loop:" in line):
                break
    coord_file.close()

    # Place everything back into the file.
    with open(metadata_file_name, 'w') as stream:
            ruamel.yaml.dump(meta_data, stream,
                             Dumper=ruamel.yaml.RoundTripDumper,
                             default_flow_style=True, indent=4,
                             canonical=False)
    stream.close()



import h5py
import math
import numpy as np
import logging

from . import tilejoining
from . import inout

""" Functions to create necessary structures and fill them in the hfd5
file.

Functions:
create_structures_hdf5_files
fill_hdf5_files
divide_final_image
fill_hdf5_MPI

"""
logger = logging.getLogger(__name__)


########################### HDF5 file creation ##########################
def create_structures_hdf5_files(stitching_file, joining, nr_pixels,
                                 z_size, hyb_nr, gene,
                                 ubyte = True, blend = 'non linear',
                                 reg_name = ''):
    """Takes an HDF5 file handle and creates the necessary structures.

    Creates groups and data sets, when the groups or data sets already
    exists, they are kept as they are, as long as the data sets have
    the right size and data type. Incompatible data sets will be
    overwritten.
    Stitching file has the following structure:

    __________________Groups____________|___Data sets____

    gene1:
          StitchedImage:
                                        final_image
                                        final_image_ubyte
                                        blending_mask
                                        temp_mask
          StitchingTempTiles:
                         blended_tiles:
                                        0
                                        1
                                        .
                                        .
                                        n
                         ubytes:
                                        0
                                        1
                                        .
                                        .
                                        n
                         temp_masks:
                                        0
                                        1
                                        .
                                        .
                                        n
   gene2: ...



    Parameters:
    -----------

    stitching_file: pointer
        HDF5 file handle. The file where the stitched images will be saved.
    joining: dict
        Dictionary containing keys 'corner_list  and 'final_image_shape'. 
        Corner_list is a list of list, each list is a pair of an image number
        (int) and it's coordinates (numpy array containing floats).
        Final_image_shape is a tuple of size 2 or 3 depending on the number of 
        dimensions and contains ints.
    nr_pixels: int
        Height and length of the tile in pixels, tile is assumed to be square.
    z_size: int 
        The number of layers in one tile (size of the z-axis). 
        Should be 1 or None if the tile is 2D.
    hyb_nr: int 
        The number of the hybridization we are stitching. 
        This will be used to place the data in the right group in stitching_file.
    gene: str
        The name of the gene we are stitching.This will be used to place the
        data in the right group in stitching_file.
    ubyte: bool 
        We need to create a data set for the ubyte image when True.
        Otherwise we only need to create a data set for the high resolution image.
    blend: str 
        When 'non linear' or 'linear', blending will be applied, so we will 
        need to create the structures necessary for saving the blended tiles. When
        it has another value or is None no blending at all will be applied, 
        so we can skip this. This variable also determines to return value of
        linear_blending.
    reg_name: str 
        A name that will be used to create a second StitchedImage group for 
        an image that is registered to be aligned with the other hybridizations. 
        warp_name will be attached to the end of the group name, which is
        StitchedImage.
    
    Returns:
    --------

    stitched_group: pointer
        HDF5 reference to the group where the final will be.
    temp_group: pointer
        HDF5 reference to the group where the blended tiles will be saved.
    linear_blending: bool
        When True later blending should be linear and when False, blending
        should be non-linear.
    ubyte: bool
        The ubyte image should be saved when True. Otherwise only the high
        resolution image should be saved.
    blend: str 
        When 'non linear' or 'linear', blending should be applied. When
        it has another value or is None no blending at all will be applied.


    """
    logger.info("Generating stitching file structures.")
    # Create a group for the stitched images in the stitching file
    #stitching_file.require_group('Hybridization' + str(hyb_nr))
    stitching_file.require_group(gene)
    stitched_group = stitching_file[gene].require_group('StitchedImage' + reg_name)
    # Create the final image in this file
    try:
        final_image = stitched_group.require_dataset('final_image',
                                                joining['final_image_shape'],
                                                dtype = np.float64)
    except TypeError as err:
        logger.info("Incompatible 'final_image' data set already existed, deleting old dataset.\n {}"
                    .format(err))
        del stitched_group['final_image']
        inout.free_hdf5_space(stitching_file)
        final_image = stitched_group.require_dataset('final_image',
                                                joining['final_image_shape'],
                                                dtype = np.float64)
    # Create a low resolution image
    if ubyte:
        try:
            final_image_ubyte = stitched_group.require_dataset('final_image_ubyte',
                                                       joining['final_image_shape'],
                                                       dtype=np.uint8)
        except TypeError as err:
            logger.info("Incompatible 'final_image_ubyte' data set already existed, deleting old dataset.\n {}"
                        .format(err))
            del stitched_group['final_image_ubyte']
            inout.free_hdf5_space(stitching_file)
            final_image = stitched_group.require_dataset('final_image_ubyte',
                                                       joining['final_image_shape'],
                                                       dtype=np.uint8)

    # If blending is required initialize the blending mask in the
    # hdf5 file
    if blend is not None:
        # For the blending masks use only the last 2 dimensions of final
        # image shape, because also when working in 3D the masks can be
        # 2D as there is the same shift in x and y direction for the
        # whole stack.
        try:
            blending_mask = stitched_group.require_dataset('blending_mask',
                                                   joining['final_image_shape'][-2:],
                                                   dtype = np.float64)
        except TypeError as err:
            logger.info("Incompatible 'blending_mask' data set already existed, deleting old dataset.\n {}"
                        .format(err))
            del stitched_group['blending_mask']
            inout.free_hdf5_space(stitching_file)
            final_image = stitched_group.require_dataset('blending_mask',
                                                   joining['final_image_shape'][-2:],
                                                   dtype = np.float64)
        try:
            temp_mask = stitched_group.require_dataset('temp_mask',
                                               joining['final_image_shape'][-2:],
                                               dtype = np.float64)
        except TypeError as err:
            logger.info("Incompatible 'temp_mask' data set already existed, deleting old dataset.\n {}"
                        .format(err))
            del stitched_group['temp_mask']
            inout.free_hdf5_space(stitching_file)
            final_image = stitched_group.require_dataset('temp_mask',
                                               joining['final_image_shape'][-2:],
                                               dtype = np.float64)


    # Check type of blending
    if blend == 'non linear':
        linear_blending = False
    elif blend == 'linear':
        linear_blending = True
    else:
        linear_blending = False
        logger.warning("Blend not defined correctly, \
                                using non-linear blending, \
                                blend is: {}".format(blend))

    # Generate the tmp_file structure
    temp_group = stitching_file[gene].require_group('StitchingTempTiles' + reg_name)
    blended_tiles = temp_group.require_group('blended_tiles')
    ubytes = temp_group.require_group('ubytes')
    temporary_masks = temp_group.require_group('temp_masks')

    for idx in range(len(joining['corner_list'])):
            if z_size == 1 or z_size is None:
                try:
                    blended_tiles.require_dataset(str(idx), (nr_pixels, nr_pixels), np.float64)
                except TypeError as err:
                    logger.debug("Incompatible data set for blended tile {}, deleting old dataset.\n {}"
                                .format(idx, err))
                    del blended_tiles[str(idx)]
                    inout.free_hdf5_space(stitching_file)
                    blended_tiles.require_dataset(str(idx), (nr_pixels, nr_pixels), np.float64)
                try:
                    ubytes.require_dataset(str(idx), (nr_pixels, nr_pixels), np.uint8)
                except TypeError as err:
                    logger.debug("Incompatible data set for ubyte tile {}, deleting old dataset.\n {}"
                                .format(idx, err))
                    del ubytes[str(idx)]
                    inout.free_hdf5_space(stitching_file)
                    ubytes.require_dataset(str(idx), (nr_pixels, nr_pixels), np.uint8)
            else:
                try:
                    blended_tiles.require_dataset(str(idx), (z_size, nr_pixels, nr_pixels), np.float64)
                except TypeError as err:
                    logger.debug("Incompatible data set for blended tile {}, deleting old dataset.\n {}"
                                .format(idx, err))
                    del blended_tiles[str(idx)]
                    inout.free_hdf5_space(stitching_file)
                    blended_tiles.require_dataset(str(idx), (z_size, nr_pixels, nr_pixels), np.float64)
                try:
                    ubytes.require_dataset(str(idx), (z_size, nr_pixels, nr_pixels), np.uint8)
                except TypeError as err:
                    logger.debug("Incompatible data set for ubyte tile {}, deleting old dataset.\n {}"
                                .format(idx, err))
                    del ubytes[str(idx)]
                    inout.free_hdf5_space(stitching_file)
                    ubytes.require_dataset(str(idx), (z_size, nr_pixels, nr_pixels), np.uint8)

            try:
                temporary_masks.require_dataset(str(idx), (nr_pixels, nr_pixels), np.float64)
            except TypeError as err:
                logger.debug("Incompatible data set for temporary masks tile {}, deleting old dataset.\n {}"
                                .format(idx, err))
                del temporary_masks[str(idx)]
                inout.free_hdf5_space(stitching_file)
                temporary_masks.require_dataset(str(idx), (nr_pixels, nr_pixels), np.float64)
    if False:
        logger.info("Flushing hdf5 file to clean up after delete operations")
        before_flush = stitching_file.id.get_filesize()
        stitching_file.flush()
        after_flush = stitching_file.id.get_filesize()
        logger.debug("Size in bytes before flush: {} after flush: {} space freed: {}".format(before_flush, after_flush, before_flush - after_flush))
    return stitched_group, temp_group, linear_blending, ubyte, blend


def fill_hdf5_files(im_file, joining, nr_pixels, ubyte, blend):
    """
    Fill the data sets in the hdf5 file with zeros or random data.

    Parameters:
    -----------

    im_file: pointer
        HDF5 reference. Reference to the StitchedImage
        group or any group or file that contains final_image
        and optionally final_image_ubyte, blending_mask and
        temp_mask.
    joining: dictionary 
        Containing the key final_image_shape.
        Final_image_shape is a tuple of size 2 or 3 depending on
        the number of dimensions and contains ints.
    nr_pixels: int
        Height and length of the tile in pixels, tile is assumed to be square.
    ubyte: bool
        The ubyte image will be filled when True.
        Otherwise only the high resolution image will be saved.
    blend: str 
        When blend is not None, blending_mask and temp_mask will be filled.
    """
    logger.info("Filling stitching file data sets.")
    # Fill the image arrays with zeros, this works for 3D
    im_file['final_image'][:]= np.zeros(joining['final_image_shape'])
    if ubyte:
        im_file['final_image_ubyte'][:] = np.empty(joining['final_image_shape'],
                                           dtype=np.uint8)
    if blend is not None:
        im_file['blending_mask'][:] = np.zeros(joining['final_image_shape'][-2:])
        tilejoining.make_mask(joining, nr_pixels, im_file['blending_mask'])
        im_file['temp_mask'][:] = np.zeros(joining['final_image_shape'][-2:])


def divide_final_image(im_shape, nr_blocks):
    """
    Divide the image in rows if it is too large to be filled at once.

    The parallel HDF5 implementation cannot handle images larger
    than 2GB. Therefore this function can divide the rows of the image
    over nr_blocks to be filled by separate cores.

    Parameters:
    -----------

    im_shape: tuple
        Shape (y, x) of the image to be filled.
    nr_blocks: int
        The number of block to divide the image in,
        typically this will be the number of available cores.

    Returns:
    --------

    y_corners: list
        List of numpy ints (numpy.int64). The y-corners of all the blocks.
    """

    logger.debug("nr_blocks {}".format(nr_blocks))
    logger.debug("im_shape {}".format(im_shape[-2]))
    y_corners = np.linspace(0, im_shape[-2], num = nr_blocks,
                        endpoint = False, dtype = int)
    logger.debug("y_corners {}".format(y_corners))
    logger.debug('Types: y_corners: {} corner element: {}'
                    .format(type(y_corners), type(y_corners[0])))
    return y_corners


def fill_hdf5_MPI(y_corners, ind, im_shape, im_file, ubyte, blend):
    """Fill the data sets in the hdf5 file with zeros, block by block.

    The parallel HDF5 implementation cannot handle images larger
    than 2GB. Therefore this function only fills the part of the
    image from the indexing value found in y_corners[ind] up to
    y_corners[ind + 1].

    Parameters:
    -----------

    y_corners: list
        List of numpy ints (numpy.int64). The y-corners of all the blocks.
    ind : int 
        The index of the corner in y_corners
    im_shape: tuple
        Shape of the final image.
    im_file: pointer 
        HDF5 file handle or group reference. The file or
        group with the datasets that need to be filled.
    ubyte: bool
        Ubyte image is filled if True
    blend: str 
        If this is None datasets needed for blending
        are left empty, for all other values they are filled.
    """
    logger.info("Filling stitching file data sets in MPI mode.")
    # Calculate the range to be filled with zeros
    y_min = y_corners[ind]
    if (ind + 1) < len(y_corners):
        y_max = y_corners[ind + 1]
    else:
        y_max = im_shape[-2]

    logger.debug("index: {} y_min: {} y_max: {} im_shape[0]: {}"
                .format(ind, y_min, y_max, im_shape[-2]))

    # Fill with zeros:
    if len(im_shape) == 3:
        im_file['final_image'][:, y_min:y_max,:] = 0.0
        if ubyte:
            im_file['final_image_ubyte'][:, y_min:y_max,:] = 0
    else:
        im_file['final_image'][y_min:y_max,:] = 0.0
        if ubyte:
            im_file['final_image_ubyte'][y_min:y_max,:] = 0
    if blend is not None:
        im_file['blending_mask'][y_min:y_max,:] = 0.0
        im_file['temp_mask'][y_min:y_max,:] = 0.0

    #DO NOT FORGET to run tilejoining.make_mask after filling the arrays!!!!


########################### HDF5 file creation ##########################
def create_structures_hdf5_stitched_ref_gene_file_npy(stitching_file, joining, nr_pixels,
                                 reference_gene, blend = 'non linear'):
    """Takes an HDF5 file handle and creates the necessary structures.

    Modification of create_structures_hdf5_files to work with .npy list of
    files

    Creates groups and data sets, when the groups or data sets already
    exists, they are kept as they are, as long as the data sets have
    the right size and data type. Incompatible data sets will be
    overwritten.
    Stitching file has the following structure:

    __________________Groups____________|___Data sets____

    gene_stitched:
          StitchedImage:
                                        final_image
                                        blending_mask


    Parameters:
    -----------

    stitching_file: pointer
        HDF5 file handle. The file where the stitched images will be saved.
    joining: dict 
        Dictionary containing keys 'corner_list  and 'final_image_shape'. 
        Corner_list is a list of list, each list is a pair of an image number
        (int) and it's coordinates (numpy array containing floats).
        Final_image_shape is a tuple of size 2 or 3
        depending on the number of dimensions and contains ints.
    nr_pixels: int 
        Height and length of the tile in pixels, tile is assumed to be square.
    reference_gene: str 
        The name of the gene we are stitching.This will be used to place the
        data in the right group in stitching_file.
     blend: str
        When 'non linear' or 'linear',blending will be applied,
        so we will need to create the structures
        necessary for saving the blended tiles. When
        it has another value or is None no blending at
        all will be applied, so we can skip this.
        This variable also determines to return value of
        linear_blending.
    
    Returns:
    --------

    stitched_group: pointer
        HDF5 reference to the group where the final will be.
    linear_blending: bool
        When True later blending should be linear and when False, blending
        should be non-linear.
    blend: str 
        When 'non linear' or 'linear', blending should be applied. When
        it has another value or is None no blending at
        all will be applied.


    """
    logger.info("Generating stitching file structures.")
    # Create a group for the stitched images in the stitching file
    
    stitching_file.require_group(reference_gene)
    stitched_group = stitching_file[reference_gene].require_group('StitchedImage')
    # Create the final image in this file
    try:
        final_image = stitched_group.require_dataset('final_image',
                                                joining['final_image_shape'],
                                                dtype = np.float64)
    except TypeError as err:
        logger.info("Incompatible 'final_image' data set already existed, deleting old dataset.\n {}"
                    .format(err))
        del stitched_group['final_image']
        inout.free_hdf5_space(stitching_file)
        final_image = stitched_group.require_dataset('final_image',
                                                joining['final_image_shape'],
                                                dtype = np.float64)
    

    # If blending is required initialize the blending mask in the
    # hdf5 file
    if blend is not None:
        # For the blending masks use only the last 2 dimensions of final
        # image shape, because also when working in 3D the masks can be
        # 2D as there is the same shift in x and y direction for the
        # whole stack.
        try:
            blending_mask = stitched_group.require_dataset('blending_mask',
                                                   joining['final_image_shape'][-2:],
                                                   dtype = np.float64)
        except TypeError as err:
            logger.info("Incompatible 'blending_mask' data set already existed, deleting old dataset.\n {}"
                        .format(err))
            del stitched_group['blending_mask']
            inout.free_hdf5_space(stitching_file)
            final_image = stitched_group.require_dataset('blending_mask',
                                                   joining['final_image_shape'][-2:],
                                                   dtype = np.float64)

    # Check type of blending
    if blend == 'non linear':
        linear_blending = False
    elif blend == 'linear':
        linear_blending = True
    else:
        linear_blending = False
        logger.warning("Blend not defined correctly, \
                                using non-linear blending, \
                                blend is: {}".format(blend))

    
    if False:
        logger.info("Flushing hdf5 file to clean up after delete operations")
        before_flush = stitching_file.id.get_filesize()
        stitching_file.flush()
        after_flush = stitching_file.id.get_filesize()
        logger.debug("Size in bytes before flush: {} after flush: {} space freed: {}".format(before_flush, after_flush, before_flush - after_flush))
    return stitched_group, linear_blending, blend


import argparse
import pickle
import numpy as np
import glob
import nd2reader as nd2
import os
import logging
import shutil

from distributed import Client
from skimage import io

from pysmFISH.counting import filtering_and_counting_experiment
from pysmFISH import utils

def process_standalone_experiment():

    """
    Script to run conversion, filtering and raw counting on a small set of images.
    The analysis run locally

    All the parameters are entered with argparse

    Parameters:
    -----------

    path: string
        Path to the experiment to process
    analysis_name: string
        Name of the analysis
    stringency: int
        Value of the stringency to use in the threshold selection. Default=0
    min_distance: int
        Min distance betwenn to peaks. Default=5
    min_plane: int
        Min plane for z-stack cropping. Default=None
    max_plane: int:
        Max plane for z-stack cropping. Default=None
    ncores: int
        Number of cores to use for the processing. Deafault=1


    """

    # input to the function
    parser = argparse.ArgumentParser(description='Counting and filtering experiment')
    parser.add_argument('-path', help='path to experiment to analyze')
    parser.add_argument('-analysis_name', help='analysis name')
    parser.add_argument('-stringency', help='stringency', default=0, type=int)
    parser.add_argument('-min_distance', help='min distance between peaks', default=5, type=int)
    parser.add_argument('-min_plane', help='starting plane to consider', default=None, type=int)
    parser.add_argument('-max_plane', help='ending plane to consider', default=None, type=int)
    parser.add_argument('-ncores', help='number of cores to use', default=1, type=int)



    # Parse the input args
    args = parser.parse_args()
    processing_directory = args.path
    analysis_name = args.analysis_name
    stringency = args.stringency
    min_distance = args.min_distance
    min_plane = args.min_plane
    max_plane = args.max_plane
    ncores  = args.ncores
                  

    if min_plane != None and max_plane != None:
        plane_keep = [min_plane, max_plane]
    else:
        plane_keep = None


    # Determine the os type
    os_windows, add_slash = utils.determine_os()

    # Starting logger
    utils.init_file_logger(processing_directory)
    logger = logging.getLogger()

    logger.debug('min_plane%s', min_plane)
    logger.debug('max_plane %s', max_plane)
    logger.debug('keep_planes value %s',plane_keep)

    # Start the distributed client
    client = Client(n_workers = ncores, threads_per_worker=1)

    logger.debug('client %s',client)
    logger.debug('check that workers are on the same directory %s', client.run(os.getcwd))
    
    # Check trail slash
    processing_directory = utils.check_trailing_slash(processing_directory,os_windows)


    # Determine the experiment name
    exp_name = processing_directory.split(add_slash)[-2]

    logger.debug('Experiment name: %s', exp_name)


    # Create the directories where to save the output
    tmp_dir_path = processing_directory+analysis_name+'_'+exp_name+'_tmp'+add_slash
    filtered_dir_path = processing_directory+analysis_name+'_'+exp_name+'_filtered'+add_slash
    counting_dir_path = processing_directory+analysis_name+'_'+exp_name+'_counting_pkl'+add_slash
    try:
        os.stat(tmp_dir_path)
    except:
        os.mkdir(tmp_dir_path)
        os.chmod(tmp_dir_path,0o777)

    try:
        os.stat(filtered_dir_path)
    except:
        os.mkdir(filtered_dir_path)
        os.chmod(filtered_dir_path,0o777)

    try:
        os.stat(counting_dir_path)
    except:
        os.mkdir(counting_dir_path)
        os.chmod(counting_dir_path,0o777)


    # Get the list of the nd2 files to process inside the directory
    files_list = glob.glob(processing_directory+'*.nd2')
    logger.debug('files to process %s', files_list)


    # Convert the .nd2 data
    for raw_data_gene_fname in files_list:
        fname = raw_data_gene_fname.split(add_slash)[-1][:-4]
        logger.debug('fname %s', fname)
        with nd2.Nd2(raw_data_gene_fname) as nd2file:
            for channel in nd2file.channels:
                for fov in nd2file.fields_of_view:
                    img_stack = np.empty([len(nd2file.z_levels),nd2file.height,nd2file.width],dtype='uint16')
                    images =  nd2file.select(channels=channel, fields_of_view=fov,z_levels=nd2file.z_levels)
                    for idx,im in enumerate(images):
                        img_stack[idx,:,:] = im

                    converted_fname=tmp_dir_path+exp_name+'_'+fname+'_'+channel+'_fov_'+str(fov)+'.npy'
                    np.save(converted_fname,img_stack,allow_pickle=False)     


    logger.debug('Finished .nd2 file conversion')

    # Filtering all the data
    # Get list of the files to process
    flist_img_to_filter=glob.glob(tmp_dir_path+'*.npy')

    # logger.debug('files to filter %s',flist_img_to_filter)
    # Parallel process all the data
    futures_processes=client.map(filtering_and_counting_experiment,flist_img_to_filter, \
                                  filtered_dir_path=filtered_dir_path, \
                                 counting_dir_path=counting_dir_path, \
                                 exp_name=exp_name,plane_keep=plane_keep,add_slash=add_slash, \
                                 min_distance=min_distance, stringency=stringency)
    
    
    client.gather(futures_processes)               
    client.close()

    logger.debug('Finished filtering and counting')         

    # delete the tmp folders
    shutil.rmtree(tmp_dir_path)

if __name__ == "__main__":
    process_standalone_experiment()



import numpy as np
import nd2reader as nd2

def nd2_to_npy(gene,raw_data_gene_fname,tmp_gene_dir,processing_hyb,use_ram=True,max_ram=150000000000):
       
    """    
    Function used to convert Nikon ND2 files in numpy arrays.
    Important: require the nd2reader (https://github.com/jimrybarski/nd2reader)
    pip  install "nd2reader==2.1.3". There is a new version based on pims
    but need to be tested
    The input ND2 file will have only one channel and will be converted in a
    binary npy.
    
    Parameters:
    -----------

    gene: str 
        Name of the gene to process (ex. Aldoc)
    raw_data_gene_fname: str
        Path to the ND2 file (ex. '/Users/simone/Desktop/test/ND2conversion/BW_Aldoc.nd2')
    tmp_gene_dir: str 
        path to the folder where to save the npy files
    processing_hyb: str 
        Name of the processing_hyb to process (ex. EXP-17-BP3597_hyb2)
    use_ram: bool 
        If True the images are written first in ram until max_ram is reached
    max_ram: int
        Size of the RAM that can be used as temporary storage
 
    """   
    
    size_counter = 0
    tmp_storage={}
    
    with nd2.Nd2(raw_data_gene_fname) as nd2file:
        for channel in nd2file.channels:
            for fov in nd2file.fields_of_view:
                img_stack = np.empty([len(nd2file.z_levels),nd2file.height,nd2file.width],dtype='uint16')
                images = nd2file.select(channels=channel, fields_of_view=fov,z_levels=nd2file.z_levels)
                for idx,im in enumerate(images):
                    img_stack[idx,:,:] = im
            
                converted_fname=tmp_gene_dir+processing_hyb+'_'+gene+'_pos_'+str(fov)+'.npy'
                
                if use_ram:
                    # Create a dict that is saved in ram. When full is written on disk
                    # Done to reduce the number of i/o and increase performance
                    size_counter += img_stack.nbytes
                    if size_counter < max_ram:
                        tmp_storage[fov]={}
                        tmp_storage[fov]['img']=img_stack
                        tmp_storage[fov]['converted_fname']=converted_fname
                    else:
                        for pos in tmp_storage.keys():
                            np.save(tmp_storage[pos]['converted_fname'],tmp_storage[pos]['img'],allow_pickle=False)

                        tmp_storage={}
                        size_counter = img_stack.nbytes
                        tmp_storage[fov]={}
                        tmp_storage[fov]['img'] = img_stack
                        tmp_storage[fov]['converted_fname'] = converted_fname                  
                
                else:
                    # Directly save the file without saving it in RAM
                    np.save(converted_fname,img_stack,allow_pickle=False)
        
        # Save the data that remained in the tmp_storage (if RAM is used) 
        if use_ram:
            for pos in tmp_storage.keys():
                        np.save(tmp_storage[pos]['converted_fname'],tmp_storage[pos]['img'],allow_pickle=False)



"""Find or apply coordinates to stitch an image
"""

#import matplotlib.pyplot as plt
#plt_available = True
plt_available = False
import sklearn.feature_extraction.image as sklim
import numpy as np
import skimage.transform as smtf
import h5py
import logging
import time
import os

# Own imports
from . import inout
from . import pairwisesingle as ps
from .MicroscopeData import MicroscopeData
from .GlobalOptimization import GlobalOptimization
from . import tilejoining

from .. import utils

# Logger
logger = logging.getLogger(__name__)

#################### Initial stitching functions #######################
def get_pairwise_input(ImageProperties,folder, tile_file, hyb_nr, gene = 'Nuclei',
                        pre_proc_level = 'FilteredData',
                        est_overlap = 0.1, y_flip = False, nr_dim = 2):
    """Get the information necessary to do the pairwise allignment

    Find the pairwise pars for an unknown stitching.
    Works best with a folder containing
    image with nuclei (DAPI staining)

    Parameters:
    -----------

    folder: str
        String representing the path of the folder containing
        the tile file and the yaml metadata file. Needs a
        trailing slash ('/').
    tile_file: pointer
        HDF5 file handle. Reference to the opened file containing the tiles.
    hyb_nr: int
        The number of the hybridization we are going to
        stitch. This will be used to navigate tile_file and find
        the correct tiles.
    gene: str
        The name of the gene we are going to stitch.
        This will be used to navigate tile_file and find the
        correct tiles. (Default: 'Nuclei')
    pre_proc_level: str
        The name of the pre processing group of
        the tiles we are going to stitch.
        This will be used to navigate tile_file and find the
        correct tiles. (Default: 'Filtered')
    est_overlap: float
        The fraction of two neighbours that should
        overlap, this is used to estimate the shape of the
        tile set and then overwritten by the actual average
        overlap according to the microscope coordinates.
        (default: 0.1)
    y_flip: bool
        The y_flip variable is designed for the cases where the
        microscope sequence is inverted in the y-direction. When
        set to True the y-coordinates will also be inverted
        before determining the tile set. (Default: False)
    nr_dim: int
        If 3, the code will assume three dimensional data
        for the tile, where z is the first dimension and y and x
        the second and third. For any other value 2-dimensional data
        is assumed. (Default: 2)
    
    Returns:
    --------

    tiles: list
        List of references to the the tiles in the hdf5 file tile_file.
    contig_tuples: list
        List of tuples. Each tuple is a tile pair.
        Tuples contain two tile indexes denoting these
        tiles are contingent to each other.
    nr_pixels: int
        Height and length of the tile in pixels, tile is assumed to be square.
    z_count: int
        The number of layers in one tile (size of the z-axis). Is 1 when nr_dim is not 3.
    micData: object
        MicroscopeData object. Contains coordinates of the tile corners as taken from the microscope.
    """
    logger.info("Getting files from folder: {}".format(folder))

    # Load the information from the metadata file.
    ExperimentInfos, ImageProperties, HybridizationsInfos, \
    Converted_Positions, MicroscopeParameters = \
        utils.experimental_metadata_parser(folder)
    # Get coordinate data for this hybridization
    coord_data = Converted_Positions['Hybridization' + str(hyb_nr)]

    # Read the number of pixels, z-count and pixel size from the yaml
    # file.
    try:
        nr_pixels       = ImageProperties['HybImageSize']['rows']
    except KeyError as err:
        logger.info(("Number of pixels not found in experimental "
                     + "metadata file.\nPlease add "
                     + "the number of pixels in an image "
                     + "to the experimental "
                     + "metadata file under ImageProperties "
                     + "--> HybImageSize --> rows.\n"
                     + "KeyError: {}").format(err))
        raise

    if nr_dim == 2:
        z_count = 1
    else:
        try:
            z_count = ImageProperties['HybImageSize']['zcount']
        except KeyError as err:
            logger.info(("Number of pixels not found in experimental "
                         + "metadata file.\nPlease add "
                         + "the number of slices in the z-stack "
                         + "to the experimental "
                         + "metadata file under ImageProperties "
                         + "--> HybImageSize --> zcount.\n"
                         + "KeyError: {}")
                         .format(err))
            raise

    try:
        pixel_size = ImageProperties['PixelSize']
    except KeyError as err:
        logger.info(("ImageProperties['PixelSize'] not found in "
                    + "experimental metadata file.\nPlease add the "
                    + "size of a pixel in um in the experimental "
                    + "metadata file under ImageProperties "
                    + "--> PixelSize.\nKeyError: {}").format(err))
        raise
    # Estimate the overlap in pixels with the overlap that the user
    # provided, default is 10%
    est_x_tol = nr_pixels * (1 - est_overlap)
    logger.info("Estimating overlap at {}%, that is {} pixels"
                .format(est_overlap * 100, est_x_tol))
    logger.debug("Number of pixels: {}".format(nr_pixels))
    logger.debug("Number of slices in z-stack: {}".format(z_count))

    # Organize the microscope data and determine tile set
    micData = MicroscopeData(coord_data, y_flip, nr_dim)
    micData.normalize_coords(pixel_size)
    micData.make_tile_set(est_x_tol, nr_pixels = nr_pixels)

    # Make a list of image numbers, matching with the numbers in the
    # image files
    flat_tile_set = micData.tile_set.flat[:]

    image_list = [micData.tile_nr[ind] if ind >= 0 else -1 for ind in flat_tile_set]
    image_list = np.ma.masked_equal(image_list, -1)
    logger.info("Getting references for: {}".format(image_list))

    # Make a list of the image names
    tiles = inout.get_image_names(tile_file, image_list = image_list,
                                    hyb_nr = hyb_nr, gene = gene,
                                    pre_proc_level = pre_proc_level)

    logger.info("Size tiles: {} Number of pixels: {} z count: {}"
                .format(len(tiles), nr_pixels, z_count))

    # Produce an undirected graph of the tiles, tiles that are
    # neighbours to each other are connected in this graph.
    # noinspection PyPep8Naming
    C = np.asarray(sklim.grid_to_graph(*micData.tile_set.shape).todense())
    np.fill_diagonal(C, 0)
    # noinspection PyPep8Naming
    C = np.triu( C )
    # Extract the neighbour pairs from the graph
    contig_tuples =list(zip( *np.where( C ) ))
    logger.info(("Length contingency tuples: {} \n"
                + "Contingency tuples: {}")
                 .format(len(contig_tuples), contig_tuples))

    # Plotting tiles:
    #inout.display_tiles(tiles, micData.tile_set, fig_nr = 2, block = False)
    #plt.show(block = True)
    return (tiles, contig_tuples, nr_pixels, z_count, micData)

def get_pairwise_alignments(tiles, tile_file, contig_tuples,
                                        micData, nr_peaks = 8,
                                        nr_slices = None,
                                        nr_dim = 2):
    """Calculate the pairwise transition

    Calculates pairwise transition for each neighbouring pair of
    tiles. This functions is only used in the single core version of the
    code, not when using MPI.

    Parameters:
    -----------

    tiles: list
        List of references to the the tiles in the hdf5 file tile_file.
    tile_file: pointer
        HDF5 file handle. Reference to the opened file containing the tiles.
    contig_tuples: list
        List of tuples. Each tuple is a tile pair.
        Tuples contain two tile indexes denoting these
        tiles are contingent to each other.
    micData: object
        MicroscopeData object. Containing coordinates of
        the tile corners as taken from the microscope.
    nr_peaks: int
        Number of peaks to be extracted from the PCM (Default: 8)
    nr_slices: int 
        Only applicable when running with 3D
        pictures and using 'compres pic' method in
        pairwisesingle.py. Determines the number of slices
        that are compressed together (compression in the
        z-direction). If None, all the slices are compressed
         together. (Default: None)
    nr_dim: int
        If 3, the code will assume three dimensional data
        for the tile, where z is the first dimension and y and x
        the second and third. For any other value 2-dimensional data
        is assumed. (default: 2)

    Returns:
    --------

    : dict
        Contains key 'P' with a 1D numpy array
        containing pairwise alignment y and x coordinates
        (and z-coordinates when applicable) for each
        neighbouring pair of tiles, array will be
        2 * len(contig_typles) for 2D data
        or 3 * len(contig_typles) for 3D data.
        Also contains key 'covs' with a 1D numpy array
        containing covariance for each pairwise alignment in
        'P', 'covs' will be len(contig_typles).
    """
    logger.info("Getting pairwise alignments...")
    P = np.empty((len(contig_tuples), nr_dim), dtype = int)
    covs = np.empty(len(contig_tuples))

    for i in range(len(contig_tuples)):
        # noinspection PyPep8Naming
        P_single, cov, contig_index = ps.align_single_pair(tiles, tile_file,
                                        contig_tuples, i,
                                        micData, nr_peaks,
                                        nr_slices = nr_slices,
                                        nr_dim = nr_dim)
        P[contig_index,:] = P_single
        covs[contig_index] = cov

    logger.info("Raw P: {}".format(P))
    #Flatten P
    P = np.array(P).flat[:]
    logger.info("flat P: {}".format(P))
    return {'P': P, 'covs': covs}

############################# Apply ####################################
def get_place_tile_input_apply(folder, tile_file, hyb_nr, data_name,
                                gene = 'Nuclei',
                                pre_proc_level = 'Filtered',
                                nr_dim = 2, check_pairwise = False):
    """Get the data needed to apply stitching to another gene

    Parameters:
    -----------

    folder: str
        String representing the path of the folder containing
        the tile file, the stitching data file the yaml metadata
        file. Needs a trailing slash ('/').
    tile_file: pointer
        HDF5 file handle. Reference to the opened file
        containing the tiles.
    hyb_nr: int
        The number of the hybridization we are going to
        stitch. This will be used to navigate tile_file and find
        the correct tiles.
    data_name: str
        Name of the file containing the pickled stitching data.
    gene: str
        The name of the gene we are going to stitch.
        This will be used to navigate tile_file and find the
        correct tiles. (Default: 'Nuclei')
    pre_proc_level: str
        The name of the pre processing group of
        the tiles we are going to stitch.
        This will be used to navigate tile_file and find the
        correct tiles. (Default: 'Filtered')
    nr_dim: int
        If 3, the code will assume three dimensional data
        for the tile, where z is the first dimension and y and x
        the second and third. For any other value 2-dimensional data
        is assumed. (Default: 2)
    check_pairwise: bool
        If True the contig_tuples array is assumed
        to be in the pickled data file and will be returned.
        (Default: False)

    Returns:
    --------

    joining: dict
        Taken from the stitching data file.
        Contains keys corner_list and final_image_shape.
        Corner_list is a list of list, each list is a pair
        of an image number (int) and it's coordinates (numpy
        array containing floats).
        Final_image_shape is a tuple of size 2 or 3
        depending on the numer of dimensions and contains
        ints.
    tiles: list
        List of references to the the tiles in the hdf5 file tile_file.
    nr_pixels: int
        Height and length of the tile in pixels, tile is assumed to be square.
    z_count: int
        The number of layers in one tile (size of
        the z-axis). Is 1 when nr_dim is not 3.
    micData: object
        MicroscopeData object. Taken from the pickled
        stitching data.
        Contains coordinates of the tile corners as taken
        from the microscope.
    contig_tuples: list
        Only returned if check_pairwise == True.
        list of tuples. Taken from the pickled
        stitching data. Each tuple is a tile pair.
        Tuples contain two tile indexes denoting these
        tiles are contingent to each other.
    """
    logger.info("Getting data to apply stitching from file...")

    # Load image list and old joining data
    stitching_coord_dict = inout.load_stitching_coord(folder + data_name)
    # noinspection PyPep8Naming
    micData = stitching_coord_dict['micData']
    joining = stitching_coord_dict['joining']
    logger.info("Joining object and image list loaded from file")

    # Make a list of image numbers
    flat_tile_set = micData.tile_set.flat[:]
    image_list = [micData.tile_nr[ind] if ind >= 0 else -1 for ind in flat_tile_set]
    image_list = np.ma.masked_equal(image_list, -1)
    logger.info("Tile set size: {}".format(micData.tile_set.shape))
    logger.info("Placing folowing image references in tiles: {}"
                .format(image_list))

    # Make a list of the tile references
    tiles = inout.get_image_names(tile_file, image_list = image_list,
                                hyb_nr = hyb_nr, gene = gene, pre_proc_level = pre_proc_level)

    # Load the data from the metadata file
    ExperimentInfos, ImageProperties, HybridizationsInfos, \
    Converted_Positions, MicroscopeParameters = \
        utils.experimental_metadata_parser(folder)

    # Read the number of pixels and z-count from the yaml
    # file.
    try:
        nr_pixels = ImageProperties['HybImageSize']['rows']
    except KeyError as err:
        logger.info(("Number of pixels not found in experimental "
                     + "metadata file.\nPlease add "
                     + "the number of pixels in an image "
                     + "to the experimental "
                     + "metadata file under ImageProperties "
                     + "--> HybImageSize --> rows.\n"
                     + "KeyError: {}").format(err))
        raise

    if nr_dim == 2:
        z_count = 1
    else:
        try:
            z_count = ImageProperties['HybImageSize']['zcount']
        except KeyError as err:
            logger.info(
                ("Number of pixels not found in experimental "
                 + "metadata file.\nPlease add "
                 + "the number of slices in the z-stack "
                 + "to the experimental "
                 + "metadata file under ImageProperties "
                 + "--> HybImageSize --> zcount.\n"
                 + "KeyError: {}")
                .format(err))
            raise

    logger.info("Size tiles: {} Number of pixels: {} z count: {}"
                .format(len( tiles), nr_pixels, z_count))

    # Check pairwise overlap in signal
    if check_pairwise:
        contig_tuples   = stitching_coord_dict['contig_tuples']
        logger.info(
            "Length contingency tuples: {} Contingency tuples: {}"
            .format(len(contig_tuples), contig_tuples))
        return (joining, tiles, nr_pixels, z_count, micData,
                                            contig_tuples)
    else:
        return (joining, tiles, nr_pixels, z_count, micData)


############################# Refine ###################################
def get_refine_pairwise_input(folder, tile_file, hyb_nr, data_name,
                                gene = 'Nuclei', pre_proc_level = 'Filtered',
                                nr_dim = 2):
    """Get the data needed to refine stitching with another gene

    Parameters:
    -----------

    folder: str
        String representing the path of the folder containing
        the tile file, the stitching data file the yaml metadata
        file. Needs a trailing slash ('/').
    tile_file: pointer
        HDF5 file handle. Reference to the opened file
        containing the tiles.
    hyb_nr: int
        The number of the hybridization we are going to
        stitch. This will be used to navigate tile_file and find
        the correct tiles.
    data_name: str
        Name of the file containing the pickled stitching data.
    gene: str
        The name of the gene we are going to stitch.
        This will be used to navigate tile_file and find the
        correct tiles. (Default: 'Nuclei')
    pre_proc_level: str
        The name of the pre processing group of
        the tiles we are going to stitch.
        This will be used to navigate tile_file and find the
        correct tiles. (Default: 'Filtered')
    nr_dim: int
        If 3, the code will assume three dimensional data
        for the tile, where z is the first dimension and y and x
        the second and third. For any other value 2-dimensional data
        is assumed. (Default: 2)

    Returns:
    --------

    tiles: list
        List of references to the the tiles in the hdf5 file tile_file.
    contig_tuples: list
        List of tuples. Each tuple is a tile pair.
        Tuples contain two tile indexes denoting these
        tiles are contingent to each other.
    nr_pixels: int
        Height and length of the tile in pixels, tile
        is assumed to be square.
    z_count: int
        The number of layers in one tile (size of the z-axis). Is 1 when nr_dim is not 3.
    micData: object
        MicroscopeData object. Contains coordinates of
        the tile corners as taken from the microscope.
    """
    logger.info("Aplying stitching from file")

    # Load image list and old joining data
    stitching_coord_dict = inout.load_stitching_coord(folder + data_name)
    # noinspection PyPep8Naming
    micData = stitching_coord_dict['micData']
    logger.info("Joining object and image list loaded from file")

    flat_tile_set = micData.tile_set.flat[:]
    image_list = [micData.tile_nr[ind] if ind >= 0 else -1 for ind in flat_tile_set]
    image_list = np.ma.masked_equal(image_list, -1)
    logger.info("Tile set size: {}".format(micData.tile_set.shape))
    logger.info("Loading images: {}".format(image_list))

    # Make a list of the image names
    tiles = inout.get_image_names(tile_file, image_list = image_list,
                                hyb_nr = hyb_nr, gene = gene, pre_proc_level = pre_proc_level)

    contig_tuples   = stitching_coord_dict['contig_tuples']
    alignment_old   = stitching_coord_dict['alignment']['P']

    # Load the data from the metadata file
    ExperimentInfos, ImageProperties, HybridizationsInfos, \
    Converted_Positions, MicroscopeParameters = \
        utils.experimental_metadata_parser(folder)

    # Read the number of pixels and z-count from the yaml
    # file.
    try:
        nr_pixels = ImageProperties['HybImageSize']['rows']
    except KeyError as err:
        logger.info(("Number of pixels not found in experimental "
                     + "metadata file.\nPlease add "
                     + "the number of pixels in an image "
                     + "to the experimental "
                     + "metadata file under ImageProperties "
                     + "--> HybImageSize --> rows.\n"
                     + "KeyError: {}").format(err))
        raise

    if nr_dim == 2:
        z_count = 1
    else:
        try:
            z_count = ImageProperties['HybImageSize']['zcount']
        except KeyError as err:
            logger.info(
                ("Number of pixels not found in experimental "
                 + "metadata file.\nPlease add "
                 + "the number of slices in the z-stack "
                 + "to the experimental "
                 + "metadata file under ImageProperties "
                 + "--> HybImageSize --> zcount.\n"
                 + "KeyError: {}")
                    .format(err))
            raise


    # Recalculate C
    C = sklim.grid_to_graph(*micData.tile_set.shape).todense()
    np.fill_diagonal(C,0)
    C = np.triu( C )

    return (tiles, contig_tuples, nr_pixels, z_count, micData, C, alignment_old)


def refine_pairwise_alignments(tiles, tile_file, contig_tuples, alignment_old,
                                        micData = None, nr_peaks = 8,
                                        nr_dim = 2):
    """Calculate the pairwise transition

    Calculates pairwise transition for each neighbouring pair of
    tiles.

    Parameters:
    -----------

    tiles: np.array
        Array of tiles, a tile should be a 2d np.array
        representing a picture
    contig_tuples: list
        List of tuples denoting which tiles are contingent to each other.
    micData: object
        MicroscopeData object containing coordinates (default None)
    nr_peaks: int
        nr of peaks to be extracted from the PCM (default 8)
    nr_dim: int
        If 3, the code will assume three dimensional data
        for the tile, where z is the first dimension and y and x
        the second and third. For any other value 2-dimensional data
        is assumed. (default: 2)

    Returns:
    --------
    : dict
        Contains key 'P' with a 1D numpy array
        containing pairwise alignment y and x coordinates
        (and z-coordinates when applicable) for each
        neighbouring pair of tiles, array will be
        2 * len(contig_typles) for 2D data
        or 3 * len(contig_typles) for 3D data.
        Also contains key 'covs' with a 1D numpy array
        containing covariance for each pairwise alignment in
        'P', 'covs' will be len(contig_typles).
    """
    # Make a new P and cov list for the refine pairwise alignments
    P_ref = np.empty((len(contig_tuples), nr_dim), dtype = int)
    covs_ref = np.empty(len(contig_tuples))

    for i in range(len(contig_tuples)):
        P_single, cov, contig_index  = ps.refine_single_pair(tiles,
                                        tile_file,
                                        contig_tuples, i, micData,
                                        alignment_old['P'], nr_peaks,
                                        nr_dim = nr_dim)
        P_ref[contig_index,:] = P_single
        covs_ref[contig_index] = cov

    logger.info("Raw P: {}".format(P_ref))
    # Flatten P
    P_ref = np.array(P_ref).flat[:]
    logger.info("flat P: {}".format(P_ref))
    return {'P': P_ref, 'covs': covs_ref}


######################### General functions ############################
def get_place_tile_input(folder, tiles, contig_tuples,
                            micData, nr_pixels, z_count, alignment,
                            data_name, nr_dim = 2, save_alignment = True):
    """Do the global alignment and get the shifted corner coordinates.

    Calculates a shift in global coordinates for each tile (global
    alignment) and then applies these shifts to the  corner coordinates
    of each tile and returns and saves these shifted corner coordinates.

    This function produces a file with stitching data in folder
    called data_name, this file includes the corner coordinates which
    can be used to apply the stitching to another gene.

    Parameters:
    -----------

    folder: str
        String representing the path of the folder containing
        the tile file and the yaml metadata file. Needs a
        trailing slash ('/').
    tiles: list
        List of strings. List of references to the the
        tiles in the hdf5 file tile_file.
    contig_tuples: list
        List of tuples. Each tuple is a tile pair.
        Tuples contain two tile indexes denoting these
        tiles are contingent to each other.
    micData: object
        MicroscopeData object. Contains coordinates of
        the tile corners as taken from the microscope.
    nr_pixels: int
        Height and length of the tile in pixels, tile is assumed to be square.
    z_count: int
        The number of layers in one tile (size of the z-axis). Is 1 when nr_dim is not 3.
    alignment: dict
        Contains key 'P' with a 1D numpy array
        containing pairwise alignment y and x coordinates
        (and z-coordinates when applicable) for each
        neighbouring pair of tiles, array will be
        2 * len(contig_typles) for 2D data
        or 3 * len(contig_typles) for 3D data.
        Also contains key 'covs' with a 1D numpy array
        containing covariance for each pairwise alignment in
        'P', 'covs' will be len(contig_typles).
    data_name: str
        Name of the file containing the pickled stitching data.
    nr_dim: int
        If 3, the code will assume three dimensional data
        for the tile, where z is the first dimension and y and x
        the second and third. For any other value 2-dimensional data
        is assumed. (default: 2)
    save_alignment: bool
        When False only the stitching
        coordinates and microscope data will be saved. When
        True also the contigency tuples and pairwise
        alignment will be saved (this is necessary if we
        want to refine the stitching later). (Default: True)

    Returns:
    --------
    joining: dict
        Contains keys corner_list and
        final_image_shape.
        Corner_list is a list of list, each list is a pair
        of a tile index (int) and it's tile's shifted
        coordinates in the final image (numpy array
        containing floats).
        Final_image_shape is a tuple of size 2 or 3
        depending on the numer of dimensions and contains
        ints.

    """
    # Perform global optimization
    logger.debug("Initializing global optimization")
    optimization = GlobalOptimization()
    logger.debug("Starting optimization, micData")
    optimization.performOptimization(micData.tile_set, contig_tuples,
                                alignment['P'], alignment['covs'],
                                len(tiles), nr_dim)


    # Stitch everything back together
    # Determine global corners
    joining = tilejoining.calc_corners_coord(tiles,
                optimization.global_trans, micData, nr_pixels, z_count)

    # Save the data to do the stitching in "data_name":
    if joining:
        if save_alignment:
            inout.save_to_file(folder + data_name,
                               joining = joining,
                               contig_tuples = contig_tuples,
                               alignment = alignment,
                               micData = micData)
        else:
            inout.save_to_file(folder + data_name,
                                micData = micData,
                                joining = joining)
    else:
        logger.warning("No results found to save: joining is empty")

    return joining

def assess_performance(micData, alignment, joining,
                        cov_signal, xcov_list, folder,
                        use_IJ_corners = False):
    """Assess the performance of the stitching

    This functions writes its the result to a file in "folder".

    Parameters:
    -----------

    micData: object
        MicroscopeData object. Contains coordinates of
        the tile corners as taken from the microscope
        and contains the tile set.
    alignment: dict
        Contains key 'P' with a 1D numpy array
        containing pairwise alignment y and x coordinates
        (and z-coordinates when applicable) for each
        neighbouring pair of tiles, array will be
        2 * len(contig_typles) for 2D data
        or 3 * len(contig_typles) for 3D data.
        Also contains key 'covs' with a 1D numpy array
        containing covariance for each pairwise alignment in
        'P', 'covs' will be len(contig_typles).
    joining: dict
        Contains keys corner_list and
        final_image_shape.
        Corner_list is a list of list, each list is a pair
        of a tile index (int) and it's tile's shifted
        coordinates in the final image (numpy array
        containing floats).
        Final_image_shape is a tuple of size 2 or 3
        depending on the numer of dimensions and contains
        ints.
    cov_signal: np.array
        The covariance of each neighbouring
        tile pair in the part of the tiles that overlap in
        the final stitched signal image.
    xcov_list: list
        List of cross covariance of the
        overlap of the tiles in the final stitched image.
        As returned by tilejoining.assess_overlap.
    folder: str
        String representing a path. The folder where the
        performance report should be saved. Needs a
        trailing slash ('/').
    use_IJ_corners: bool
        If True compare our corners to Image J
        found in a file in folder, the file name should
        contain: TileConfiguration.
    """
    ################################Gather data#########################
    report_string   = ""
    if use_IJ_corners:
        compare_corners = inout.read_IJ_corners(folder)
        # Make a list of image numbers, matching with the numbers in the
        # image files
        flat_tile_set = micData.tile_set.flat[:]
        image_list = [micData.tile_nr[ind] if ind >= 0
                      else -1 for ind in flat_tile_set]
        image_list = np.ma.masked_equal(image_list, -1)
        # Compare our corners to the corner the ImageJ plugin found
        # Select the tiles we actually used:
        compare_corners_new = [[item[0], np.array([item[1][1], item[1][0]])]
                                for item in compare_corners
                                if item[0] in image_list]
        # Replace the indexes with numbering of the images:
        logger.debug('My corners {}'.format(joining['corner_list']))
        my_corners_new = [[image_list[i], item[1]] for i, item in enumerate(joining['corner_list'])]
        logger.debug('My corners new {}'.format(my_corners_new))
        logger.debug('Compare corners new {}'.format(compare_corners_new))
        # Normalize, to make the first one the origin and compare
        compare_origin  = compare_corners_new[0][1]
        my_origin       = my_corners_new[0][1]
        logger.debug("origins: {}, {}".format(compare_origin, my_origin))

        cum_diff = np.zeros((1,2))
        for i in range(len(my_corners_new)):
            my_cur = (my_corners_new[i][1] - my_origin)
            compare_cur = (compare_corners_new[i][1] - compare_origin)
            diff = abs(my_cur - compare_cur)
            cum_diff += diff
            report_string += ("My tile: {}, {}, compare tile: {}, {}; difference: {}\n"
                                .format(my_corners_new[i][0], my_cur,
                                        compare_corners_new[i][0], compare_cur,
                                        diff))
        report_string += "\nAverage: {}\n".format(cum_diff / len(my_corners_new))

    # Calculate average cross covariances of the overlaps in the final image:
    if xcov_list is not None:
        av_xcov = np.mean(xcov_list)
    else:
        logger.info('No cross covariance data available')
        av_xcov = None
    report_string += "\nAverage cross covariance of final overlap: {}\n".format(av_xcov)
    report_string += "Cross covariance list of final overlap: {}\n".format(xcov_list)
    #logger.debug(report_string)

    ###################### Save the performance data ###################
    perf_path = folder + 'performance/'
    # To print the logging to a file
    try:
            os.stat(perf_path)
    except:
            os.mkdir(perf_path)
            os.chmod(perf_path,0o777)

    dateTag = time.strftime("%y%m%d_%H_%M_%S")
    with open(perf_path + dateTag + '-performance' + '.txt', 'w') as f:
        f.write(report_string)
        if micData is not None:
            # If available print the alignment results:
            f.write(("\nTile set: \n{} \n"
                    "Tile numbers: {}\n")
                    .format(micData.tile_set, micData.tile_nr))
        if alignment is not None:
            # If available print the alignment results:
            f.write(("Pairwise Alignment: {}\n"
                    "Covariances: {}\n"
                    "Average covariance: {}\n")
                    .format(alignment['P'], alignment['covs'],
                            np.nanmean(alignment['covs'])))
        if joining is not None:
            f.write(("Corners after alignment: \n{}\n")
                    .format(joining['corner_list']))
        if cov_signal is not None:
            f.write(("\nAverage pairwise covariance of the signal: {}\n"
                     "Pairwise covariance of the signal:\n{}\n")
                                            .format(np.nanmean(cov_signal), cov_signal))
    f.close()


############################# Visualization ############################
def save_as_tiff(data_file, hyb_nr, gene, location_image,
                    pre_proc_level = 'StitchedImage', mode = 'both'):
    """Save the results as a tiff image for visual inspection.

    Parameters:
    -----------

    data_file: pointer
        HDF5 file handle. HDF5 file containing the final image.
    gene: str
        The name of the gene we stitched.
        This will be used to navigate data_file and find the
        correct final picture.
    hyb_nr: int
        The number of the hybridization we have
        stitched.This will be used to navigate data_file and
        find the correct final picture.
    location_image: str
        Full path to the file where the tiff file
        will be saved (extension not necessary).
    pre_proc_level: str
        The name of the pre processing group of
        the tiles we are going to stitch. Normally this will
        be 'StitchedImage', but when the final image is
        found in another datagroup it may be changed.
        This will be used to navigate data_file and find the
        correct final image. (Default: 'StitchedImage')
    mode: str
        Mode determines what color, quality and
        how many images are saved.
        Possible values for mode: save_ubyte, save_float,
        save_rgb. If another or no value is given the image
        is saved as is and a as a low quality copy
        (pixel depth 8 bits) (Default: 'both')
    """
    # Save the results:
    if mode == 'save_ubyte':
        inout.save_image(data_file, hyb_nr, gene, pre_proc_level,  'final_image_ubyte', location_image + '_byte')
    elif mode == 'save_float':
        inout.save_image(data_file, hyb_nr, gene, pre_proc_level, 'final_image', location_image)
    elif mode == 'save_rgb':
        inout.save_image(data_file, hyb_nr, gene, pre_proc_level, 'final_image_rgb', location_image + '_rgb')
    else:
        inout.save_image(data_file, hyb_nr, gene, pre_proc_level, 'final_image_ubyte', location_image + '_byte')
        inout.save_image(data_file, hyb_nr, gene, pre_proc_level, 'final_image', location_image)


def plot_final_image(im_file_name, joining,  hyb_nr = 1,
                     gene = 'Nuclei', fig_name = "final image",
                     shrink_image = False, block = True):
    """Displays the high quality final image in a plot window.

    Takes a lot of working memory for full sized images.
    When plt_available is false this function does nothing and returns
    None.

    Parameters:
    -----------

    im_file_name: str
        Filename of the hdf5 file, containing the final image.
    fig_name: str
        Name of the plotting window (default: "final image").
    shrink_image: bool
        Turn on shrink_image to reduce display quality and memory usage. (Default: False)
    block: bool
        Plot blocks the running program untill
        the plotting window is closed if true. Turn off
        block to make the code continue untill the next call
        of  plt.show(block=True) before displaying the
        image. (default: True)
    """
    if plt_available:
        if isinstance(im_file_name, str):
            # Load the image from file
            im_file = h5py.File(im_file_name + '_Hybridization' +
                                str(hyb_nr) + '.sf.hdf5', 'r')
            for_display = im_file['final_image']
        else:
            # Load the image from file
            for_display = im_file_name[gene] \
                                    ['StitchedImage']['final_image']
        # Shrink the image if necessary
        if shrink_image:

            display_size = np.array(joining['final_image_shape'],
                                            dtype = int)/10
            logger.debug("display size pixels: {}".format(display_size))
            for_display = smtf.resize(for_display, tuple(display_size))
        # Plot the image
        if for_display.ndim == 3:
            inout.plot_3D(for_display)
        else:
            plt.figure(fig_name)
            plt.imshow(for_display, 'gray', interpolation = 'none')
            plt.show(block = False)

        # Load the image from file
        if isinstance(im_file_name, str):
            # Load the image from file
            im_file = h5py.File(im_file_name + '.hdf5', 'r')
            for_display = im_file['temp_mask']
        else:
            for_display = im_file_name['Hybridization' + str(hyb_nr)][gene] \
                        ['StitchedImage']['temp_mask']
        # Shrink the image if necessary
        if shrink_image:
            display_size = np.array(joining.final_image_shape, dtype=int) / 10
            logger.debug("display size pixels: {}".format(display_size))
            for_display = smtf.resize(for_display, tuple(display_size))
        # Plot the image
        plt.figure(fig_name + ' mask')
        plt.imshow(for_display, 'gray', interpolation='none')
        plt.show(block = block)
    else:
        return None


def get_pairwise_input_npy(image_properties,converted_positions, hybridization,
                        est_overlap = 0.1, y_flip = False, nr_dim = 2):
    """Get the information necessary to do the pairwise allignment
    Modified version of the get_pairwise_input functions that work on .npy 
    files and not on hdf5

    Find the pairwise pairs for an unknown stitching.
    
    Parameters:
    -----------
    
    image_properties: dict 
        Dictionary with the image details parsed from the Experimental_metadata.yaml file
    converted_positions: dict
        Dictionary  with the coords of the images for all hybridization
        The coords are a list of floats
    hybridization: str 
        Hybridization that will be processed (Ex. Hybridization2)
    est_overlap: float
        The fraction of two neighbours that should
        overlap, this is used to estimate the shape of the
        tile set and then overwritten by the actual average
        overlap according to the microscope coordinates.
        (default: 0.1)
    y_flip: bool
        The y_flip variable is designed for the cases where the
        microscope sequence is inverted in the y-direction. When
        set to True the y-coordinates will also be inverted
        before determining the tile set. (Default: False)
    nr_dim: int
        If 3, the code will assume three dimensional data
        for the tile, where z is the first dimension and y and x
        the second and third. For any other value 2-dimensional data
        is assumed. (Default: 2)
    
    Returns:
    --------

    tiles: np.array
        Array of int with the tiles number. -1 indicate an empty tile
    contig_tuples: list
        List of tuples. Each tuple is a tile pair.
        Tuples contain two tile indexes denoting these
        tiles are contingent to each other.
    nr_pixels: int
        Height and length of the tile in pixels, tile is assumed to be square.
    z_count: int
        The number of layers in one tile (size of
        the z-axis). Is 1 when nr_dim is not 3.
    micData: object
        MicroscopeData object. Contains coordinates of
        the tile corners as taken from the microscope.
    """


    # Get coordinate data for this hybridization
    coord_data = converted_positions[hybridization]

    # Read the number of pixels, z-count and pixel size from the yaml
    # file.
    try:
        nr_pixels       = image_properties['HybImageSize']['rows']
    except KeyError as err:
        logger.info(("Number of pixels not found in experimental "
                     + "metadata file.\nPlease add "
                     + "the number of pixels in an image "
                     + "to the experimental "
                     + "metadata file under ImageProperties "
                     + "--> HybImageSize --> rows.\n"
                     + "KeyError: {}").format(err))
        raise

    if nr_dim == 2:
        z_count = 1
    else:
        try:
            z_count = image_properties['HybImageSize']['zcount']
        except KeyError as err:
            logger.info(("Number of pixels not found in experimental "
                         + "metadata file.\nPlease add "
                         + "the number of slices in the z-stack "
                         + "to the experimental "
                         + "metadata file under ImageProperties "
                         + "--> HybImageSize --> zcount.\n"
                         + "KeyError: {}")
                         .format(err))
            raise

    try:
        pixel_size = image_properties['PixelSize']
    except KeyError as err:
        logger.info(("ImageProperties['PixelSize'] not found in "
                    + "experimental metadata file.\nPlease add the "
                    + "size of a pixel in um in the experimental "
                    + "metadata file under ImageProperties "
                    + "--> PixelSize.\nKeyError: {}").format(err))
        raise
    # Estimate the overlap in pixels with the overlap that the user
    # provided, default is 10%
    est_x_tol = nr_pixels * (1 - est_overlap)
    logger.info("Estimating overlap at {}%, that is {} pixels"
                .format(est_overlap * 100, est_x_tol))
    logger.debug("Number of pixels: {}".format(nr_pixels))
    logger.debug("Number of slices in z-stack: {}".format(z_count))

    # Organize the microscope data and determine tile set
    micData = MicroscopeData(coord_data, y_flip, nr_dim)
    micData.normalize_coords(pixel_size)
    micData.make_tile_set(est_x_tol, nr_pixels = nr_pixels)

    # Make a list of image numbers, matching with the numbers in the
    # image files
    flat_tile_set = micData.tile_set.flat[:]

    image_list = [micData.tile_nr[ind] if ind >= 0 else -1 for ind in flat_tile_set]
    image_list = np.ma.masked_equal(image_list, -1)
    logger.info("Getting references for: {}".format(image_list))


    # Make a list of the image names (-1 is a missing tile)
    tiles = image_list.data


    # Produce an undirected graph of the tiles, tiles that are
    # neighbours to each other are connected in this graph.
    # noinspection PyPep8Naming
    C = np.asarray(sklim.grid_to_graph(*micData.tile_set.shape).todense())
    np.fill_diagonal(C, 0)
    # noinspection PyPep8Naming
    C = np.triu( C )
    # Extract the neighbour pairs from the graph
    contig_tuples =list(zip( *np.where( C ) ))
    logger.info(("Length contingency tuples: {} \n"
                + "Contingency tuples: {}")
                 .format(len(contig_tuples), contig_tuples))

    return(tiles, contig_tuples, nr_pixels, z_count, micData)


def get_place_tile_input_apply_npy(hyb_dir,stitched_reference_files_dir,data_name,image_properties,nr_dim=2):
    """
    Modified version of the get_place_tile_input_apply
    Get the data needed to apply stitching to another gene

    Parameters:
    -----------

    hyb_dir: str
        String representing the path of the folder containing
        the tile file, the stitching data file the yaml metadata file.
    stitched_reference_files_dir: str
        String representing the path of the folder containing the registered data.
    data_name: str  
        Name of the file containing the pickled stitching data.
    image_properties: dict
        Dictionary with the image details parsed from the Experimental_metadata.yaml file
    nr_dim: int
        If 3, the code will assume three dimensional data
        for the tile, where z is the first dimension and y and x
        the second and third. For any other value 2-dimensional data
        is assumed. (Default: 2)
    
    Returns:
    --------

    joining: dict
        Taken from the stitching data file.
        Contains keys corner_list and final_image_shape.
        Corner_list is a list of list, each list is a pair
        of an image number (int) and it's coordinates (numpy
        array containing floats).
        Final_image_shape is a tuple of size 2 or 3
        depending on the numer of dimensions and contains
        ints.
    tiles: list
        List of strings. List of references to the the tiles in the hdf5 file tile_file.
    nr_pixels: int
        Height and length of the tile in pixels, tile is assumed to be square.
    z_count: int
        The number of layers in one tile (size of the z-axis). Is 1 when nr_dim is not 3.
    micData: object
        MicroscopeData object. Taken from the pickled stitching data.
        Contains coordinates of the tile corners as taken from the microscope.
    """
    
    logger.info("Getting data to apply stitching from file...")

    # Load image list and old joining data
    stitching_coord_dict = inout.load_stitching_coord(stitched_reference_files_dir + data_name)
    # noinspection PyPep8Naming
    micData = stitching_coord_dict['micData']
    joining = stitching_coord_dict['joining']
    logger.info("Joining object and image list loaded from file")
    
    # Make a list of image numbers
    flat_tile_set = micData.tile_set.flat[:]
    image_list = [micData.tile_nr[ind] if ind >= 0 else -1 for ind in flat_tile_set]
    image_list = np.ma.masked_equal(image_list, -1)
    logger.info("Tile set size: {}".format(micData.tile_set.shape))
    logger.info("Placing folowing image references in tiles: {}"
                .format(image_list))
    
    # Make a list of the image names (-1 is a missing tile)
    tiles = image_list.data
    
    
    # Read the number of pixels and z-count from the yaml
    # file.
    try:
        nr_pixels = image_properties['HybImageSize']['rows']
    except KeyError as err:
        logger.info(("Number of pixels not found in experimental "
                     + "metadata file.\nPlease add "
                     + "the number of pixels in an image "
                     + "to the experimental "
                     + "metadata file under ImageProperties "
                     + "--> HybImageSize --> rows.\n"
                     + "KeyError: {}").format(err))
        raise

    if nr_dim == 2:
        z_count = 1
    else:
        try:
            z_count = image_properties['HybImageSize']['zcount']
        except KeyError as err:
            logger.info(
                ("Number of pixels not found in experimental "
                 + "metadata file.\nPlease add "
                 + "the number of slices in the z-stack "
                 + "to the experimental "
                 + "metadata file under ImageProperties "
                 + "--> HybImageSize --> zcount.\n"
                 + "KeyError: {}")
                .format(err))
            raise

    logger.info("Size tiles: {} Number of pixels: {} z count: {}"
                .format(len( tiles), nr_pixels, z_count))


    return (joining, tiles, nr_pixels, z_count, micData)



"""Functions to perform registration between all hybridizations.

register_final_images(folder, gene='Nuclei',
                      sub_pic_frac=0.2, use_MPI=False,
                      apply_to_corners=True, apply_warping = True)
        -- Register stitched images an in all  HDF5 file in the folder
find_reg_final_image(im_file_1, im_file_n,
                           max_trans, sub_pic_frac,
                           nr_peaks=8)
        -- Find the transform that registers image n correctly onto
        image 1.
transform_final_image(im_file_n, trans, new_size)
        -- Transform an image according to trans.
transform_data_file(folder, data_name, trans,
                        new_size)
        -- Transform the corners in the pickled data file
align_sub_region(overlap1, overlap2, nr_peaks)
        -- Determine how much overlap2 should be shifted to fit
        overlap1, help function for find_reg_final_image
"""

import numpy as np
import h5py
import os
import skimage.transform as smtf
try:
    from mpi4py import MPI
    MPI_available = True
except ImportError:
    MPI_available = False
import logging
import glob

# Own imports
from . import inout
from . import pairwisesingle as ps

logger = logging.getLogger(__name__)


def register_final_images(folder, gene='Nuclei',
                          sub_pic_frac=0.2, use_MPI=False,
                          apply_to_corners=True, apply_warping = False,
                          region=None, compare_in_seq=False):
    """Register stitched images an in all  HDF5 file in the folder

    Loops the hybridizations in the HDF5 file, takes the stitched
    images as indicated by gene and then compares each image to the
    first image.
    For the comparison only a small patch of the images is used, the
    size of this patch can be controlled with "sub_pic_frac".

    Parameters:
    -----------

    folder: str
        The name of the folder containing the pickled file with stitching data, 
        needs a trailing slash ("/").
    gene: str 
        The gene of which the stitched images are present and should be realigned.
        Typically this will be 'Nuclei', because the smFISH genes will not have 
        enough signal to align the pictures properly. (Default: 'Nuclei')
    sub_pic_frac: float 
        The fraction of the size of the original image that should be used to compare
        images. (Default: 0.2)
    use_MPI: bool 
        If True open the files in MPI friendly mode, if False open files in normal
        single processing mode. (Default: False)
    apply_to_corners: bool
        Determines if the found registration will be applied to the tile
        corners in the pickled stitching data file. (Default: True)
    apply_warping: bool
        Determines if the found registration will be applied as a warp to the
        final pictures in the hdf5 file, should not be used with large datasets.
        (Default: False)
    region: list 
        List of length four containing ints. The region that should be compared to determine
        the shift needed for registration. Should be in the order: [y_min, y_max, x_min,
        x_max]. When region is defined, sub_pic_frac will not be used.
        By default the code will determine the region itself taking a area around the
        center of the image with a size determined by sub_pic_frac(Default: None)
    compare_in_seq: bool 
        Determines if we should compare images in sequence or if we should compare
        all to the first image.

    """
    if not compare_in_seq:
        file_name_list, file_1, im_file_1, trans, old_size_list, \
            max_trans = \
            prepare_for_comparing(folder, gene, compare_in_seq,
                                  use_MPI=use_MPI)
        # Compare each file to file 1:
        for i in range(1, len(file_name_list)):
            cur_trans, max_trans, cur_old_size, file_ind = \
                get_single_trans(file_name_list, i, gene, im_file_1,
                                 max_trans, sub_pic_frac=sub_pic_frac,
                                 region=region, use_MPI=use_MPI)
            trans[file_ind, :] = cur_trans
            old_size_list[file_ind, :] = cur_old_size
        # Close the hdf5 file.
        file_1.close()
        trans, new_size = correct_trans_and_size(trans,
                                                 old_size_list,
                                                 max_trans,
                                                 compare_in_seq)
    else:
        file_name_list, trans_relative, old_size_list, max_trans = \
            prepare_for_comparing(folder, gene, compare_in_seq,
                                  use_MPI=use_MPI)
        # Compare each file to previous file:
        for i in range(1, len(file_name_list)):
            cur_trans, max_trans, cur_old_size, file_ind = \
                get_single_relative_trans(file_name_list, i, gene,
                                          max_trans,
                                          sub_pic_frac = sub_pic_frac,
                                          region = region,
                                          use_MPI = use_MPI)
            trans_relative[file_ind, :] = cur_trans
            old_size_list[file_ind, :]  = cur_old_size
        trans, new_size = correct_trans_and_size(trans_relative,
                                                 old_size_list,
                                                 max_trans,
                                                 compare_in_seq)

    logger.debug(
        'Files: {} Translations: {}'
            .format(file_name_list, trans))

    # Apply the translations
    for i in range(len(file_name_list)):
        if apply_warping:
            if use_MPI:
                file_n = h5py.File(file_name_list[i], 'r+',
                                   driver='mpio', comm=MPI.COMM_WORLD)
            else:
                file_n = h5py.File(file_name_list[i], 'r+')
            im_file_n = file_n[gene]['StitchedImage']
            transform_final_image(im_file_n, trans[i, :], new_size)
            file_n.close()
        if apply_to_corners:
            data_name = (
            os.path.split(file_name_list[i])[1].split(sep='.')[0]
            + '_' + gene
            + '_stitching_data')
            transform_data_file(folder, data_name, trans[i, :],
                                new_size)

def prepare_for_comparing(folder, gene, compare_in_seq, use_MPI=False):
    """
    Prepare the file list, first file and init other lists.

    Parameters:
    -----------

    folder: str         
        The name of the folder containing the
        pickled file with stitching data, needs a
        trailing slash ("/").
    gene: str
        The gene of which the stitched
        images are present and should be realigned.
        Typically this will be 'Nuclei', because the
        smFISH genes will not have enough signal to
        align the pictures properly.
        (Default: 'Nuclei')
    compare_in_seq: bool
        Determines if we should compare
        images in sequence or if we should compare
        all to the first image.
    use_MPI: bool
        If True open the files in MPI
        friendly mode, if False open files in normal
        single processing mode. (Default: False)
    
    Returns:
    --------
    file_name_list: list
        List of strings. List of the sf.hdf5-files in the folder.
    trans: np.array
        Array of ints. The array to store the translations, initialized with zeros.
    old_size_list: np.array 
        Array of ints. The array to store the sizes of the final images, initialized with
        zeros.
    max_trans: np.array
        Array of ints. Variable to store the
        largest translation found up to now,
        initialized at zero.
    
    Notes:
    ------
    Only returned if compare_in_seq is True:
    file_1: pointer
        File handle to the first hdf5 file in the folder.
    im_file_1: pointer
        Reference to the group in   the first file that contains th final image.
    """
    # Get a list of files in the folder
    file_name_list = glob.glob(folder + '*.sf.hdf5')
    file_name_list.sort()
    logger.debug('Filenames sorted: {}'.format(file_name_list))
    # Initialize some variables:
    trans = np.zeros((len(file_name_list), 2), dtype=int)
    old_size_list = np.zeros((len(file_name_list), 2), dtype=int)
    max_trans = np.zeros((1, 2), dtype=int)

    # Take the first hybridization (keys() seems to give the groups
    # as a sorted list)
    im_name_1 = file_name_list[0]
    logger.debug('im_name_1: {}'.format(im_name_1))
    # Open the stitching file and make a list of the hybridizations
    # present in this file:
    if use_MPI:
        file_1 = h5py.File(im_name_1, 'r+',
                           driver='mpio', comm=MPI.COMM_WORLD)
    else:
        file_1 = h5py.File(im_name_1, 'r+')

    # hyb_name_list = list(stitching_file.keys())

    # Get the right group
    im_file_1 = file_1[gene]['StitchedImage']
    # Get the size of the first image in the list,
    # which will be the reference image without translation.
    old_size_list[0, :] = im_file_1['final_image'].shape
    # Make comparisons
    if not compare_in_seq:
        return file_name_list, file_1, im_file_1, trans, \
               old_size_list, max_trans
    else:
        # Get the size for the first image
        file_1.close()
        return file_name_list, trans, old_size_list, max_trans


def get_single_trans(file_name_list, i, gene,
                     im_file_1, max_trans, sub_pic_frac=0.2,
                     region=None, use_MPI=False):
    """Get the translation between image 1 and image i.

    Get the translation between the image in file 1 and file i
    from file_name_list.

    Parameters:
    -----------

    file_name_list: list
        List of strings. List of the sf.hdf5-files in the folder.
    i: int
        Index of the current file to compare.
    gene: str
        Gene of which the stitched
        images are present and should be realigned.
        Typically this will be 'Nuclei', because the
        smFISH genes will not have enough signal to
        align the pictures properly.
        (Default: 'Nuclei')
    im_file_1: pointer
        Reference to the group in the first file that contains th final image.
    max_trans: np.array
        Variable to store the largest translation found up to now,
        initialized at zero.
    sub_pic_frac: float
        The fraction of the size of the original image that should be used to compare
        images. (Default: 0.2)
    region: list
        List of length four containing ints. The
        region that should be compared to determine
        the shift needed for registration.
        Should be in the order: [y_min, y_max, x_min,
        x_max]. When region is defined, sub_pic_frac
        will not be used.
        By default the code will determine the region
        itself taking a area around the
        center of the image with a size
        determined by sub_pic_frac(Default: None)
    use_MPI: bool
        If True open the files in MPI friendly mode, if False open files in normal
        single processing mode. (Default: False)

    Returns:
    --------

    cur_trans: np.array
        Array of ints. Translation found between the two images that are currently
        being compared.
    max_trans: np.array 
        Array of ints. The largest translation found up to now.
    cur_old_size: np.array
        Array of ints. The sizes of the original final image found in file_name_list
        at index i.
    i: int
        The index of the second image file used for the current comparison (The first image
        file is file 1).
    """
    # Get the group containing the image we want to compare with.
    if use_MPI:
        file_n = h5py.File(file_name_list[i], 'r+',
                           driver='mpio', comm=MPI.COMM_WORLD)
    else:
        file_n = h5py.File(file_name_list[i], 'r+')
    im_file_n = file_n[gene]['StitchedImage']
    # Find the translation
    cur_trans, max_trans, cur_old_size \
        = find_reg_final_image(im_file_1, im_file_n, max_trans,
                               sub_pic_frac, region=region)
    logger.debug(
        "max_trans: {}".format(max_trans))
    file_n.close()
    return cur_trans, max_trans, cur_old_size, i

def get_single_relative_trans(file_name_list, i, gene, max_trans,
                              sub_pic_frac=0.2, region=None,
                              use_MPI=False):
    """Get the translation between image i - 1 and image i.

    Get the translation between the image in file_name_list[i - 1] and
    file_name_list[i].

    Parameters:
    -----------

    file_name_list: list
        List of strings. List of the sf.hdf5-files in the folder.
    i: int
        Index of the second image in the current comparison.
    gene: str
        The gene of which the stitched
        images are present and should be realigned.
        Typically this will be 'Nuclei', because the
        smFISH genes will not have enough signal to
        align the pictures properly.
        (Default: 'Nuclei')
    max_trans: np.array
        Array of ints. Variable to store the
        largest translation found up to now,
        initialized at zero.
    sub_pic_frac: float
        The fraction of the size of the original image that should be used to compare
        images. (Default: 0.2)
    region: list
        List of length four containing ints. The
        region that should be compared to determine
        the shift needed for registration.
        Should be in the order: [y_min, y_max, x_min,
        x_max]. When region is defined, sub_pic_frac
        will not be used.
        By default the code will determine the region
        itself taking a area around the
        center of the image with a size
        determined by sub_pic_frac (Default: None)
    use_MPI: bool
        If True open the files in MPI
        friendly mode, if False open files in normal
        single processing mode. (Default: False)

    Returns:
    --------

    cur_trans: np.array
        Array of ints. Translation found between the two images that are currently
        being compared.
    max_trans: np.array
        Array of ints. The largest translation found up to now.
    cur_old_size: np.array
        The sizes of the original final image found in file_name_list
        at index i.
    i: int
        The index of the second image file used for the current comparison (The first image
        file is file 1).

    """
    # Get the group containing the image we want to compare with.
    if use_MPI:
        file_1 = h5py.File(file_name_list[i - 1], 'r+',
                           driver='mpio', comm=MPI.COMM_WORLD)
        file_2 = h5py.File(file_name_list[i], 'r+',
                           driver='mpio', comm=MPI.COMM_WORLD)
    else:
        file_1 = h5py.File(file_name_list[i - 1], 'r+')
        file_2 = h5py.File(file_name_list[i], 'r+')
    im_file_1 = file_1[gene]['StitchedImage']
    im_file_2 = file_2[gene]['StitchedImage']
    # Find the translation
    cur_trans, max_trans, cur_old_size \
        = find_reg_final_image(im_file_1, im_file_2, max_trans,
                               sub_pic_frac, region=region)
    logger.debug("max_trans: {}".format(max_trans))
    file_1.close()
    file_2.close()
    return cur_trans, max_trans, cur_old_size, i


def correct_trans_and_size(trans_relative, old_size_list, max_trans,
                           compare_in_seq):
    """Correct the translations and the size of the registered images.

    Parameters:
    -----------

    trans_relative: np.array
        Array of ints. The array with the non-corrected translation.
    old_size_list: np.array
        Array of ints. The array with the sizes of all the final,
        non-registered images.
    max_trans: np.array
        Array of ints. Variable to store the largest translation found up to now.
    compare_in_seq: bool
        Determines if we should compare images in sequence or if we should compare
        all to the first image.

    Returns:
    --------

    trans: np.array
        Array of ints. The array with the corrected translations for each image.
    new_size: np.array 
        Array of length 2 containing ints. The size the images should have after registration.
    """
    if compare_in_seq:
        # Get the normalized transistions
        trans = np.cumsum(trans_relative, axis=0)
        max_trans = np.amax(trans, axis=0)
        logger.debug(("Comparing in sequence: relative translations: "
                      + "\n {} \n normalized translations: \n{}\n"
                      .format(trans_relative, trans)))
        logger.debug("max_trans: {}".format(max_trans))
    else:
        trans = trans_relative

    # Correct translations
    trans -= max_trans
    logger.debug('old_size_list: {}'
                 .format(old_size_list))
    # Determine final image size:
    new_size_list = old_size_list + abs(trans)
    new_size = np.amax(new_size_list, axis=0)
    logger.debug('new_size_list: {}  new_size: {}'
                 .format(new_size_list, new_size))
    return trans, new_size


def find_reg_final_image(im_file_1, im_file_n, max_trans, sub_pic_frac,
                         region=None, nr_peaks=8):
    """
    Find the transform that registers image n correctly onto image 1.
    
    Parameters:
    im_file_1: pointer
        HDF5 group reference or file handle, should
        contain a dataset "final_image" holding image 1.
    im_name_n: pointer
        HDF5 group reference or file handle, should
        contain a dataset "final_image" holding image n.
    max_trans: np.array
        Array of length 2 with dtype: int.
        Largest translation currently found.
    sub_pic_frac: float 
        The fraction of the size of the original 
        image that should be used to compare images.
    region: list
        List of length four containing ints. The
        region that should be compared to determine
        the shift needed for registration.
        Should be in the order: [y_min, y_max, x_min,
        x_max]. When region is defined, sub_pic_frac
        will not be used.
        By default the code will determine the region
        itself taking a area around the
        center of the image with a size
        determined by sub_pic_frac(Default: None)
    nr_peaks: int        
        The number of peaks used to get the best peaks 
        from the phase correlation matrix. (default: 8)

    Returns:
    --------

    trans: np.array
        Array of length 2 containing ints.
        Translation that projects image n correctly onto image 1.
    max_trans: np.array
        Array of shape (1, 2) containing ints.
        The max_trans value that was passed to this
        function, replaced by (part of) the current
        translation if it is larger than max_trans.
    shape_n: tuple
        Tuple of python ints. The shape of image n.
    """
    # Get the image shapes
    shape_1 = im_file_1['final_image'].shape
    shape_n = im_file_n['final_image'].shape

    if region is None:
        # Determine the size of the part of the picture that we want to compare
        sub_pic_size = (np.array(shape_1) * sub_pic_frac).astype(int,
                                                                 copy=False)
        logger.debug('sub_pic_size: {}'.format(sub_pic_size))
        # Take the center coordinates in the y and x axes
        center = (int(np.floor(min(shape_1[-2] / 2,
                                   shape_n[-2] / 2))),
                  int(np.floor(min(shape_1[-1] / 2,
                                   shape_n[-1] / 2))))
        start = np.array([center[0], center[1]])
        end = np.array([min(start[-2] + sub_pic_size[-2],
                            shape_1[-2],
                            shape_n[-2]),
                        min(start[-1] + sub_pic_size[-1],
                            shape_1[-1],
                            shape_n[-1])])
    else:
        start   = np.array([region[0],region[2]])
        end     = np.array([region[1],region[3]])
        logger.debug("Area based on given region: Start: {} "
                     "End: {}".format(start, end))
    # Get the region to compare from the pictures
    if im_file_1['final_image'].ndim == 3:
        pic_1 = np.amax(im_file_1['final_image'][:, start[0]:end[0],
                        start[1]:end[1]])
    else:
        pic_1 = im_file_1['final_image'][start[0]:end[0],
                start[1]:end[1]]
    if im_file_1['final_image'].ndim == 3:
        pic_n = np.amax(im_file_n['final_image'][:, start[0]:end[0],
                        start[1]:end[1]])
    else:
        pic_n = im_file_n['final_image'][start[0]:end[0],
                start[1]:end[1]]
    # Find the best translation
    trans, best_cov = align_sub_region(pic_1, pic_n, nr_peaks)
    logger.debug('Found trans: {} \n best covariance: {}'
                 .format(trans, best_cov))
    # Adjust max trans if necessary
    max_trans = np.maximum(max_trans, np.array(trans))

    return trans, max_trans, shape_n


def transform_final_image(im_file_n, trans, new_size):
    """
    Transform an image according to trans.
    
    Parameters:
    -----------

    im_file_n: pointer 
        HDF5 group reference or file handle, should
        contain a dataset "final_image" holding image n.
    trans: np.array
        Array of len 2 containing ints. y and x transform of the image.
    new_size: tuple
        Tuple of length 2. The size of the image after the transform.
    """
    # Make the trans matrix
    trans_matrix = np.eye(3)
    trans_matrix[1][2] = trans[0]
    trans_matrix[0][2] = trans[1]

    # Make a separate dataset for the registered image.
    logger.debug('new_size {}'.format(new_size))
    try:
        registered_image = im_file_n.require_dataset('reg_image',
                                             shape=tuple(new_size),
                                             dtype=np.float64)
    except TypeError as err:
        logger.debug(
            ("Incompatible data set for reg_image, deleting old " +
             "dataset. N.B: Not cleaning up space. \n {}")
            .format(err))
        del im_file_n['reg_image']
        registered_image = im_file_n.require_dataset('reg_image',
                                                 shape=tuple(new_size),
                                                 dtype=np.float64)

    # Transform the image
    registered_image[:, :] = smtf.warp(im_file_n['final_image'],
                                   trans_matrix,
                                   output_shape=new_size, order=0)


def transform_data_file(folder, data_name, trans,
                        new_size):
    """
    Transform the corners in the pickled data file

    Parameters:
    -----------

    folder: str
        The name of the folder containing the
        pickled file with stitching data, needs a
        trailing slash ("/").
    data_name: str
        Name of the pickled file with the corner coordinates.
    trans: np.array
        Array of len 2 containing ints. y and x transform of the image.
    new_size: tuple
        Tuple of length 2. The size of the image after the transform.
    """
    # Determine the name to safe the new pickled data file.
    exp_name = '_'.join(data_name.split('_')[:-2])

    # Get the original coordinates
    loaded_data = inout.load_stitching_coord(folder + data_name)
    micData = loaded_data['micData']
    joining_original = loaded_data['joining']
    joining_new = {}
    # Translate the corners
    temp_corner_list = [[tile_ind, (corner - trans)]
                        for tile_ind, corner in
                        joining_original['corner_list']]
    logger.debug(
        'temp_corner_list: {} trans: {}'.format(temp_corner_list,
                                                trans))
    # Place the corners in the joining dictionary.
    joining_new['corner_list'] = temp_corner_list
    # Change final image shape of original:
    joining_new['final_image_shape'] = new_size

    # Save to a new file
    inout.save_to_file(folder + exp_name + '_stitching_data_reg',
                       micData=micData, joining=joining_new)


def align_sub_region(overlap1, overlap2, nr_peaks):
    """Determine how much overlap2 should be shifted to fit overlap1.
    
    Parameters:
    -----------

    overlap1: np.array
        2D numpy array. Patch of the image that should be compared.
    overlap2: np.array
        2D numpy array. Patch of the image that should be compared.
    nr_peaks: int
        The number of peaks used to get the best peaks from the phase correlation matrix.
    
    Returns:
    --------

    best_trans: np.array
        Array of len 2 containing ints. Transform that projects overlap2 
        correctly onto overlap1.
    best_cov: float
        The normalized covariance
    """

    plot_order = np.ones((1, 2))
    # Calculate possible translations
    unr_pos_transistions = ps.calculate_pos_shifts(
        overlap1, overlap2, nr_peaks, 2)
    logger.debug("Possible translations: {}".
                 format(unr_pos_transistions))
    # Do correlation over the found shifts:
    best_trans, best_cov = ps.find_best_trans(unr_pos_transistions,
                                              overlap1, overlap2,
                                              plot_order)
    # Give some feedback
    logger.info(
        "Best shift: {} covariance: {}".format(best_trans, best_cov))
    return np.array(best_trans,dtype='int16'), best_cov


def register_final_images_old(folder, gene='Nuclei',
                              sub_pic_frac=0.2, use_MPI=False,
                              apply_to_corners=True,
                              apply_warping=False,
                              region=None, compare_in_seq=False):
    """Register stitched images an in all  HDF5 file in the folder

    Loops the hybridizations in the HDF5 file, takes the stitched
    images as indicated by gene and then compares each image to the
    first image.
    For the comparison only a small patch of the images is used, the
    size of this patch can be controlled with "sub_pic_frac".

    Parameters:
    -----------

    folder: str
        The name of the folder containing the
        pickled file with stitching data, needs a
        trailing slash ("/").
    gene: str
        The gene of which the stitched
        images are present and should be realigned.
        Typically this will be 'Nuclei', because the
        smFISH genes will not have enough signal to
        align the pictures properly.
        (Default: 'Nuclei')
    sub_pic_frac: float
        The fraction of the size of the
        original image that should be used to compare
        images.
        (Default: 0.2)
    use_MPI: bool
        True open the files in MPI friendly mode, if False open files in normal
        single processing mode. (Default: False)
    apply_to_corners: bool
        Determines if the found
        registration will be applied to the tile
        corners in the pickled stitching data file.
        (Default: True)
    apply_warping: bool
        Determines if the found
        registration will be applied as a warp to the
        final pictures in the hdf5 file, should not
        be used with large datasets.
        (Default: False)
    region: list
        List of length four containing ints. The
        region that should be compared to determine
        the shift needed for registration.
        Should be in the order: [y_min, y_max, x_min,
        x_max]. When region is defined, sub_pic_frac
        will not be used.
        By default the code will determine the region
        itself taking a area around the
        center of the image with a size
        determined by sub_pic_frac(Default: None)
    compare_in_seq: bool
        Determines if we should compare
        images in sequence or if we should compare
        all to the first image.

    """
    # Get a list of files in the folder
    file_name_list = glob.glob(folder + '*.sf.hdf5')
    file_name_list.sort()
    logger.debug('Filenames sorted: {}'.format(file_name_list))
    # Initialize some variables:
    trans = np.zeros((len(file_name_list), 2), dtype=int)
    old_size_list = np.zeros((len(file_name_list), 2), dtype=int)
    max_trans = np.zeros((1, 2), dtype=int)
    # Make comparisons
    if not compare_in_seq:
        # Take the first hybridization (keys() seems to give the groups
        # as a sorted list)
        im_name_1 = file_name_list[0]
        logger.debug('im_name_1: {}'.format(im_name_1))
        # Open the stitching file and make a list of the hybridizations
        # present in this file:
        if use_MPI:
            file_1 = h5py.File(im_name_1, 'r+',
                               driver='mpio', comm=MPI.COMM_WORLD)
        else:
            file_1 = h5py.File(im_name_1, 'r+')

        # hyb_name_list = list(stitching_file.keys())

        # Get the right group
        im_file_1 = file_1[gene]['StitchedImage']
        old_size_list[0, :] = im_file_1['final_image'].shape
        # Compare each file to file 1:
        for i in range(1, len(file_name_list)):
            # Get the group containing the image we want to compare with.
            if use_MPI:
                file_n = h5py.File(file_name_list[i], 'r+',
                                   driver='mpio', comm=MPI.COMM_WORLD)
            else:
                file_n = h5py.File(file_name_list[i], 'r+')
            im_file_n = file_n[gene]['StitchedImage']
            # Find the translation
            trans[i, :], max_trans, old_size_list[i, :] \
                = find_reg_final_image(im_file_1, im_file_n, max_trans,
                                       sub_pic_frac, region=region)
            logger.debug(
                "max_trans: {}".format(max_trans))
            file_n.close()
        # Close the hdf5 file.
        file_1.close()
    else:
        # Init specific array
        trans_relative = np.zeros((len(file_name_list), 2), dtype=int)
        # Compare each file to previous file:
        for i in range(1, len(file_name_list)):
            # Get the group containing the image we want to compare with.
            if use_MPI:
                file_1 = h5py.File(file_name_list[i - 1], 'r+',
                                   driver='mpio', comm=MPI.COMM_WORLD)
                file_2 = h5py.File(file_name_list[i], 'r+',
                                   driver='mpio', comm=MPI.COMM_WORLD)
            else:
                file_1 = h5py.File(file_name_list[i - 1], 'r+')
                file_2 = h5py.File(file_name_list[i], 'r+')
            im_file_1 = file_1[gene]['StitchedImage']
            im_file_2 = file_2[gene]['StitchedImage']
            # Get the size of the first image in the list,
            # which will be the reference image without translation.
            if (i - 1) == 0:
                old_size_list[0, :] = im_file_1['final_image'].shape
            # Find the translation
            trans_relative[i, :], max_trans, old_size_list[i, :] \
                = find_reg_final_image(im_file_1, im_file_2, max_trans,
                                       sub_pic_frac, region=region)
            logger.debug("max_trans: {}".format(max_trans))
            file_1.close()
            file_2.close()
        # Get the normalized transistions
        trans = np.cumsum(trans_relative, axis=0)
        max_trans = np.amax(trans, axis=0)
        logger.debug(("Comparing in sequence: relative translations: "
                      + "\n {} \n normalized translations: \n{}\n"
                      .format(trans_relative, trans)))
        logger.debug("max_trans: {}".format(max_trans))

    # Correct translations
    trans -= max_trans
    logger.debug('old_size_list: {}'
                 .format(old_size_list))
    # Determine final image size:
    new_size_list = old_size_list + abs(trans)
    new_size = np.amax(new_size_list, axis=0)
    logger.debug(
        'Files: {} Translations: {} new_size_list: {} new_size: {}'
            .format(file_name_list, trans, new_size_list, new_size))
    # Apply the translations
    for i in range(len(file_name_list)):
        if apply_warping:
            if use_MPI:
                file_n = h5py.File(file_name_list[i], 'r+',
                                   driver='mpio', comm=MPI.COMM_WORLD)
            else:
                file_n = h5py.File(file_name_list[i], 'r+')
            im_file_n = file_n[gene]['StitchedImage']
            transform_final_image(im_file_n, trans[i, :], new_size)
            file_n.close()
        if apply_to_corners:
            data_name = (
                os.path.split(file_name_list[i])[1].split(sep='.')[0]
                + '_' + gene
                + '_stitching_data')
            transform_data_file(folder, data_name, trans[i, :],
                                new_size)


def register_final_images_reg_data_only(folder, gene='Nuclei',
                          sub_pic_frac=0.2, use_MPI=False,
                          apply_to_corners=True, apply_warping = False,
                          region=None, compare_in_seq=False):
    """Register stitched images an in all  HDF5 file in the folders. 
    It is modified from register_final_images and saves only the reg_data
    file with the new coords and nothing in the hdf5 file.

    Loops the hybridizations in the HDF5 file, takes the stitched
    images as indicated by gene and then compares each image to the
    first image.
    For the comparison only a small patch of the images is used, the
    size of this patch can be controlled with "sub_pic_frac".

    Parameters:
    -----------

    folder: str
        The name of the folder containing the
        pickled file with stitching data, needs a
        trailing slash ("/").
    gene: str
        The gene of which the stitched
        images are present and should be realigned.
        Typically this will be 'Nuclei', because the
        smFISH genes will not have enough signal to
        align the pictures properly.
        (Default: 'Nuclei')
    sub_pic_frac: float
        The fraction of the size of the original image that should be used to compare
        images. (Default: 0.2)
    use_MPI: bool
        If True open the files in MPI friendly mode, if False open files in normal
        single processing mode. (Default: False)
    apply_to_corners: bool
        Determines if the found
        registration will be applied to the tile
        corners in the pickled stitching data file.
        (Default: True)
    apply_warping: bool
        Determines if the found
        registration will be applied as a warp to the
        final pictures in the hdf5 file, should not
        be used with large datasets.
        (Default: False)
    region: list
        List of length four containing ints. The
        region that should be compared to determine
        the shift needed for registration.
        Should be in the order: [y_min, y_max, x_min,
        x_max]. When region is defined, sub_pic_frac
        will not be used.
        By default the code will determine the region
        itself taking a area around the
        center of the image with a size
        determined by sub_pic_frac(Default: None)
    compare_in_seq: bool
        Determines if we should compare images in sequence or if we should compare
        all to the first image.

    """
    if not compare_in_seq:
        file_name_list, file_1, im_file_1, trans, old_size_list, \
            max_trans = \
            prepare_for_comparing(folder, gene, compare_in_seq,
                                  use_MPI=use_MPI)
        # Compare each file to file 1:
        for i in range(1, len(file_name_list)):
            cur_trans, max_trans, cur_old_size, file_ind = \
                get_single_trans(file_name_list, i, gene, im_file_1,
                                 max_trans, sub_pic_frac=sub_pic_frac,
                                 region=region, use_MPI=use_MPI)
            trans[file_ind, :] = cur_trans
            old_size_list[file_ind, :] = cur_old_size
        # Close the hdf5 file.
        file_1.close()
        trans, new_size = correct_trans_and_size(trans,
                                                 old_size_list,
                                                 max_trans,
                                                 compare_in_seq)
    else:
        file_name_list, trans_relative, old_size_list, max_trans = \
            prepare_for_comparing(folder, gene, compare_in_seq,
                                  use_MPI=use_MPI)
        # Compare each file to previous file:
        for i in range(1, len(file_name_list)):
            cur_trans, max_trans, cur_old_size, file_ind = \
                get_single_relative_trans(file_name_list, i, gene,
                                          max_trans,
                                          sub_pic_frac = sub_pic_frac,
                                          region = region,
                                          use_MPI = use_MPI)
            trans_relative[file_ind, :] = cur_trans
            old_size_list[file_ind, :]  = cur_old_size
        trans, new_size = correct_trans_and_size(trans_relative,
                                                 old_size_list,
                                                 max_trans,
                                                 compare_in_seq)

    logger.debug(
        'Files: {} Translations: {}'
            .format(file_name_list, trans))

    # Apply the translations
    for i in range(len(file_name_list)):
        if apply_to_corners:
            data_name = (
            os.path.split(file_name_list[i])[1].split(sep='.')[0]
            + '_' + gene
            + '_stitching_data')
            transform_data_file(folder, data_name, trans[i, :],
                                new_size)